(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Check_Monad
imports Error_Monad
begin

text {*A check is either successful or fails with some error.*}
type_synonym
  'e check = "'e + unit"

abbreviation succeed :: "'e check" where
  "succeed \<equiv> return ()"

definition check :: "bool \<Rightarrow> 'e \<Rightarrow> 'e check" where
  "check b e = (if b then succeed else error e)"

lemma isOK_check [simp]:
  "isOK (check b e) = b" by (simp add: check_def isOK_def)

definition check_return :: "'a check \<Rightarrow> 'b \<Rightarrow> 'a + 'b" where
  "check_return chk res = (chk \<guillemotright> return res)"

lemma check_return [simp]:
  "check_return chk res = return res' \<longleftrightarrow> isOK (chk) \<and> res' = res"
  unfolding check_return_def by (cases chk) auto

lemma [code_unfold]:
  "check_return chk res = (case chk of Inr _ \<Rightarrow> Inr res | Inl e \<Rightarrow> Inl e)"
  unfolding check_return_def bind_def ..

lemma then_return_eq [simp]:
  "(p \<guillemotright> q) = return f \<longleftrightarrow> isOK p \<and> q = return f"
  by (cases p) simp_all

abbreviation
  check_allm :: "('a \<Rightarrow> 'e check) \<Rightarrow> 'a list \<Rightarrow> 'e check"
where
  "check_allm f xs \<equiv> forallM f xs <+? snd"

lemma isOK_check_allm [simp]:
  "isOK (check_allm f xs) \<longleftrightarrow> (\<forall>x\<in>set xs. isOK (f x))"
  by simp

abbreviation
  check_allm_index :: "('a \<Rightarrow> nat \<Rightarrow> 'e check) \<Rightarrow> 'a list \<Rightarrow> 'e check"
where
  "check_allm_index f xs \<equiv> forallM_index f xs <+? snd"

lemma isOK_check_allm_index [simp]:
  "isOK (check_allm_index f xs) \<longleftrightarrow> (\<forall>i<length xs. isOK (f (xs ! i) i))"
  by simp

abbreviation
  check_all :: "('a \<Rightarrow> bool) \<Rightarrow> 'a list \<Rightarrow> 'a check"
where
  "check_all f xs \<equiv> check_allm (\<lambda>x. if f x then succeed else error x) xs"

lemma isOK_check_all [simp]:
  "isOK (check_all f xs) = (\<forall>x\<in>set xs. f x)" by simp

abbreviation
  check_all_index :: "('a \<Rightarrow> nat \<Rightarrow> bool) \<Rightarrow> 'a list \<Rightarrow> ('a \<times> nat) check"
where
  "check_all_index f xs \<equiv>
    check_allm_index (\<lambda>x i. if f x i then succeed else error (x, i)) xs"

lemma isOK_check_all_index [simp]:
  "isOK (check_all_index f xs) \<longleftrightarrow> (\<forall>i<length xs. f (xs ! i) i)" by simp

text {*The following version allows to modify the index during the check.*}
definition
  check_allm_gen_index ::
    "('a \<Rightarrow> nat \<Rightarrow> nat) \<Rightarrow> ('a \<Rightarrow> nat \<Rightarrow> 'e check) \<Rightarrow> nat \<Rightarrow> 'a list \<Rightarrow> 'e check"
where
  "check_allm_gen_index g f n xs = snd (foldl (\<lambda>(i, m) x. (g x i, m \<guillemotright> f x i)) (n, succeed) xs)"

lemma foldl_error:
  "snd (foldl (\<lambda>(i, m) x . (g x i, m \<guillemotright> f x i)) (n, error e) xs) = error e"
  by (induct xs arbitrary: n) auto

lemma isOK_check_allm_gen_index [simp]:
  assumes "isOK (check_allm_gen_index g f n xs)"
  shows "\<forall>x\<in>set xs. \<exists>i. isOK (f x i)"
using assms
proof (induct xs arbitrary: n)
  case Nil thus ?case by simp
next
  case (Cons x xs)
  show ?case
  proof (cases "isOK (f x n)")
    case True
    hence "\<exists>i. isOK (f x i)" by auto
    with True Cons show ?thesis unfolding check_allm_gen_index_def by auto
  next
    case False
    then obtain e where "f x n = error e" by (cases "f x n") auto
    with foldl_error [of g f _ e] and Cons show ?thesis
      unfolding check_allm_gen_index_def by auto
  qed
qed

lemma check_allm_gen_index [fundef_cong]:
  fixes f :: "'a \<Rightarrow> nat \<Rightarrow> 'e check"
  assumes "\<And>x n. x \<in> set xs \<Longrightarrow> g x n = g' x n"
    and "\<And>x n. x \<in> set xs \<Longrightarrow> f x n = f' x n"
  shows "check_allm_gen_index g f n xs = check_allm_gen_index g' f' n xs"
proof -
  {
    fix n m
    from assms
      have "foldl (\<lambda>(i, m) x. (g x i, m \<guillemotright> f x i)) (n, m) xs
        = foldl (\<lambda>(i, m) x. (g' x i, m \<guillemotright> f' x i)) (n, m) xs"
      by (induct xs arbitrary: n m) auto
  }
  thus ?thesis unfolding check_allm_gen_index_def by simp
qed

definition
  check_subseteq :: "'a list \<Rightarrow> 'a list \<Rightarrow> 'a check"
where
  "check_subseteq xs ys = check_all (\<lambda>x. x \<in> set ys) xs"

lemma isOK_check_subseteq [simp]:
  "isOK (check_subseteq xs ys) \<longleftrightarrow> set xs \<subseteq> set ys"
  by (auto simp: check_subseteq_def)

definition
  check_same_set :: "'a list \<Rightarrow> 'a list \<Rightarrow> 'a check"
where
  "check_same_set xs ys = (check_subseteq xs ys \<guillemotright> check_subseteq ys xs)"

lemma isOK_check_same_set [simp]:
  "isOK (check_same_set xs ys) \<longleftrightarrow> set xs = set ys"
  unfolding check_same_set_def by auto

definition
  check_disjoint :: "'a list \<Rightarrow> 'a list \<Rightarrow> 'a check"
where
  "check_disjoint xs ys = check_all (\<lambda>x. x \<notin> set ys) xs"

lemma isOK_check_disjoint [simp]:
  "isOK (check_disjoint xs ys) \<longleftrightarrow> set xs \<inter> set ys = {}"
  unfolding check_disjoint_def by (auto)

definition
  check_all_combinations :: "('a \<Rightarrow> 'a \<Rightarrow> 'b check) \<Rightarrow> 'a list \<Rightarrow> 'b check"
where
  "check_all_combinations c xs = check_allm (\<lambda>x. check_allm (c x) xs) xs"

lemma isOK_check_all_combinations [simp]:
  "isOK (check_all_combinations c xs) \<longleftrightarrow>
    (\<forall>x\<in>set xs. \<forall>y\<in>set xs. isOK (c x y))"
  unfolding check_all_combinations_def by simp

fun
  check_pairwise :: "('a \<Rightarrow> 'a \<Rightarrow> 'b check) \<Rightarrow> 'a list \<Rightarrow> 'b check"
where
  "check_pairwise c [] = succeed" |
  "check_pairwise c (x # xs) = (check_allm (c x) xs \<guillemotright> check_pairwise c xs)"

lemma pairwise_aux:
  "(\<forall>j<length (x # xs). \<forall>i<j. P ((x # xs) ! i) ((x # xs) ! j))
     = ((\<forall>j<length xs. P x (xs ! j)) \<and> (\<forall>j<length xs. \<forall>i<j. P (xs ! i) (xs ! j)))"
  (is "?C = (?A \<and> ?B)")
proof (intro iffI conjI)
  assume "?A \<and> ?B" show "?C"
  proof (intro allI impI)
    fix i j assume "j < length (x # xs)" and "i < j"
    thus "P ((x # xs) ! i) ((x # xs) ! j)"
    proof (induct j)
      case (Suc j) thus ?case
      proof (induct i)
        case 0 thus ?case using `?A \<and> ?B` by simp
      next
        case (Suc i) thus ?case using `?A \<and> ?B` by force
      qed
    qed simp
  qed
qed force+

lemma isOK_check_pairwise [simp]:
  "isOK (check_pairwise c xs) \<longleftrightarrow> (\<forall>j<length xs. \<forall>i<j. isOK (c (xs ! i) (xs ! j)))"
proof (induct xs)
  case (Cons x xs)
  note IH = this
  have "isOK (check_allm (c x) xs) = (\<forall>j<length xs. isOK (c x (xs ! j)))"
    using all_set_conv_all_nth[of xs "\<lambda>y. isOK (c x y)"] by simp
  hence "isOK (check_pairwise c (x # xs)) =
    ((\<forall>j<length xs. isOK (c x (xs ! j))) \<and>
    (\<forall>j<length xs. \<forall>i<j. isOK (c (xs ! i) (xs ! j))))" by (simp add: IH)
  thus ?case using pairwise_aux[of x xs "\<lambda>x y. isOK (c x y)"] by simp
qed simp

end

