(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Tree_Automata 
imports
  Term_Impl
  Trs
  Check_Monad
begin

subsection {* Basic definitions and Lemmas *}

datatype ('q,'f)ta_rule = TA_rule 'f "'q list" 'q ("_ _ \<rightarrow> _")

record ('q,'f)ta = 
  ta_final :: "'q set"
  ta_rules :: "('q,'f) ta_rule set"
  ta_eps   :: "('q \<times> 'q)set"

(* use prefix r for rule *)
fun r_sym where "r_sym (f qs \<rightarrow> q) = (f, length qs)"
fun r_lhs_states where "r_lhs_states (f qs \<rightarrow> q) = qs"
fun r_rhs where "r_rhs (f qs \<rightarrow> q) = q"
definition r_states where "r_states ta_rule = insert (r_rhs ta_rule) (set (r_lhs_states ta_rule))"

definition ta_states :: "('q,'f)ta \<Rightarrow> 'q set" where 
  "ta_states TA \<equiv> \<Union> (r_states ` ta_rules TA) \<union> \<Union> ((\<lambda> (q,q'). {q,q'}) ` ta_eps TA) \<union> ta_final TA"

definition ta_rhs_states :: "('q,'f)ta \<Rightarrow> 'q set"
where "ta_rhs_states TA \<equiv> {q'. \<exists> q. (q \<in> r_rhs ` ta_rules TA) \<and> (q,q') \<in> (ta_eps TA)^*}"

lemma ta_rhs_states_subset_states: "ta_rhs_states TA \<subseteq> ta_states TA"
proof
  fix q
  assume "q \<in> ta_rhs_states TA" 
  from this[unfolded ta_rhs_states_def]
  obtain q' where q': "q' \<in> r_rhs ` ta_rules TA" and qq': "(q', q) \<in> (ta_eps TA)\<^sup>*" by blast
  from q' have q': "q' \<in> ta_states TA" unfolding ta_states_def by (auto simp: r_states_def)
  from qq' have "q' = q \<or> (\<exists> q''. (q'', q) \<in> ta_eps TA)" by (metis rtranclE)
  thus "q \<in> ta_states TA" using q' unfolding ta_states_def by blast
qed

definition ta_syms :: "('q,'f)ta \<Rightarrow> 'f sig"
where "ta_syms TA \<equiv> r_sym ` ta_rules TA"

lemma finite_ta_syms: assumes "finite (ta_rules TA)" shows "finite (ta_syms TA)"
  unfolding ta_syms_def using assms by auto

(* the reachable states of some term: ta_res, variables in terms encode states *)
fun ta_res :: "('q,'f)ta \<Rightarrow> ('f,'q)term \<Rightarrow> 'q set"
where "ta_res TA (Var q) = {q'. (q,q') \<in> (ta_eps TA)^*}"
  | "ta_res TA (Fun f ts) = {q' | q' q qs. 
    (f qs \<rightarrow> q) \<in> ta_rules TA \<and> 
    (q,q') \<in> ((ta_eps TA)^*) \<and> 
    length qs = length ts \<and> 
    (\<forall> i < length ts. qs ! i \<in> (map (ta_res TA) ts) ! i)}"

declare ta_res.simps[code del]

(* the language of an automaton *)
definition ta_lang :: "('q,'f)ta \<Rightarrow> ('f,'q)term set"
where "ta_lang TA = {t. ground t \<and> ta_final TA \<inter> ta_res TA t \<noteq> {}}"

(* is an automaton deterministic: no epsilon rules, no overlapping rules *)
definition ta_det :: "('q,'f)ta \<Rightarrow> bool" where
  "ta_det TA \<equiv> ta_eps TA = {} \<and> 
    (\<forall> f qs q q'. (f qs \<rightarrow> q) \<in> ta_rules TA \<longrightarrow> (f qs \<rightarrow> q') \<in> ta_rules TA \<longrightarrow> q = q')"

(* determinism implies unique results *)
lemma ta_detE[elim, consumes 1]: assumes det: "ta_det TA"
  shows "q \<in> ta_res TA t \<Longrightarrow> q' \<in> ta_res TA t \<Longrightarrow> q = q'"
proof -
  note det = det[unfolded ta_det_def]
  from det have [simp]: "ta_eps TA = {}" by simp
  show "q \<in> ta_res TA t \<Longrightarrow> q' \<in> ta_res TA t \<Longrightarrow> q = q'"
  proof (induct t arbitrary: q q')
    case (Var x q q')
    thus ?case by simp
  next
    case (Fun f ts q1 q2)
    from Fun(2)
    obtain qs1 where rule1: "(f qs1 \<rightarrow> q1) \<in> ta_rules TA"
      and len1: "length qs1 = length ts"
      and rec1: "\<And> i. i < length ts \<Longrightarrow> qs1 ! i \<in> ta_res TA (ts ! i)" by auto
    from Fun(3)
      obtain qs2 where rule2: "(f qs2 \<rightarrow> q2) \<in> ta_rules TA"
      and len2: "length qs2 = length ts"
      and rec2: "\<And> i. i < length ts \<Longrightarrow> qs2 ! i \<in> ta_res TA (ts ! i)" by auto
    {
      fix i
      assume i: "i < length ts"
      hence mem: "ts ! i \<in> set ts" by auto
      from Fun(1)[OF this rec1 rec2] have "qs1 ! i = qs2 ! i" using i by auto
    }
    with len1 have "qs1 = qs2" unfolding len2[symmetric] unfolding list_eq_iff_nth_eq by auto
    with rule2 have rule2: "(f qs1 \<rightarrow> q2) \<in> ta_rules TA" by auto
    from rule1 rule2 det show "q1 = q2" by auto
  qed
qed

(* we can always add epsilon transitions at the end *)
lemma ta_res_eps: assumes q: "q \<in> ta_res TA t" and q': "(q,q') \<in> (ta_eps TA)^*"
  shows "q' \<in> ta_res TA t"
  using q
proof (induct t)
  case (Var q'')
  from this[simplified] q' 
  show ?case by simp
next
  case (Fun f ts)
  from Fun(2)
  obtain q'' qs where rule: "(f qs \<rightarrow> q'') \<in> ta_rules TA"
    and len: "length qs = length ts"
    and rec: "\<And>i. i < length ts \<Longrightarrow> qs ! i \<in> ta_res TA (ts ! i)"
    and q: "(q'',q) \<in> (ta_eps TA)^*"
    by auto
  from q q' have q': "(q'',q') \<in> (ta_eps TA)^*" by auto
  with rule len rec show ?case by auto
qed
  
(* states reachable for some t \<cdot> \<sigma> *)
lemma ta_res_subst: assumes 
  x: "\<And> x. x \<in> vars_term t \<Longrightarrow> \<tau> x \<in> ta_res TA (\<sigma> x)"
  and q: "q \<in> ta_res TA (map_vars \<tau> t)"
  shows "q \<in> ta_res TA (t \<cdot> \<sigma>)"
using x q
proof (induct t arbitrary: q)
  case (Var x)
  from Var(1)[of x] Var(2) ta_res_eps
  show ?case by simp
next
  case (Fun f ts)
  from Fun(3)
  obtain q' qs where rule: "(f qs \<rightarrow> q') \<in> ta_rules TA"
    and len: "length qs = length ts"
    and rec: "\<And>i. i < length ts \<Longrightarrow> qs ! i \<in> ta_res TA (map_vars \<tau> (ts ! i))"
    and q: "(q',q) \<in> (ta_eps TA)^*"
    by auto
  let ?ts = "map (\<lambda> t. t \<cdot> \<sigma>) ts"
  {
    fix i
    assume i: "i < length ts" 
    from i have mem: "ts ! i \<in> set ts" by auto
    hence "vars_term (ts ! i) \<subseteq> vars_term (Fun f ts)" by auto 
    with Fun(1)[OF mem Fun(2) rec[OF i]] have "qs ! i \<in> ta_res TA (ts ! i \<cdot> \<sigma>)"
      by auto
  } note main = this
  have id: "Fun f ts \<cdot> \<sigma> = Fun f ?ts" by simp
  show ?case
    by (unfold id ta_res.simps, rule, rule exI[of _ q], rule exI[of _ q'], rule exI[of _ qs], intro conjI, simp, rule rule, rule q, simp add: len, simp add: main)
qed

(* decomposing substitutions: linear case *)
lemma ta_res_subst_linear: 
  assumes l: "linear_term t"
  and q: "q \<in> ta_res TA (t \<cdot> \<sigma>)"
  shows "\<exists> \<tau>. q \<in> ta_res TA (map_vars \<tau> t) \<and> (\<forall> x \<in> vars_term t. \<tau> x \<in> ta_res TA (\<sigma> x))"
using l q
proof (induct t arbitrary: q)
  case (Var x)
  from Var(2) show ?case by auto
next
  case (Fun f ts)
  from Fun(3)
  obtain q' qs where rule: "(f qs \<rightarrow> q') \<in> ta_rules TA"
    and len: "length qs = length ts"
    and rec: "\<And>i. i < length ts \<Longrightarrow> qs ! i \<in> ta_res TA (ts ! i \<cdot> \<sigma>)"
    and q: "(q',q) \<in> (ta_eps TA)^*"
    by auto
  {
    fix i
    assume i: "i < length ts" 
    from i have mem: "ts ! i \<in> set ts" by auto
    from Fun(2) mem have "linear_term (ts ! i)" by auto
    note Fun(1)[OF mem this rec[OF i]]
  }
  hence "\<forall> i. \<exists> \<tau>. i < length ts \<longrightarrow> qs ! i \<in> ta_res TA (map_vars \<tau> (ts ! i)) \<and>
             (\<forall> x \<in> vars_term (ts ! i). \<tau> x \<in> ta_res TA (\<sigma> x))"
    by blast
  from choice[OF this] obtain \<tau> where
    ind: "\<And> i. i < length ts \<Longrightarrow> qs ! i \<in> ta_res TA (map_vars (\<tau> i) (ts ! i)) \<and> 
             (\<forall> x \<in> vars_term (ts ! i). (\<tau> i) x \<in> ta_res TA (\<sigma> x))"
    by auto
  from Fun(2) have "is_partition (map vars_term ts)" by auto
  from subst_merge[OF this, of \<tau>] obtain \<tau>2 where \<tau>2: "\<And> i. i < length ts \<Longrightarrow> \<forall> x \<in> vars_term (ts ! i). \<tau>2 x = \<tau> i x" by auto
  show ?case
  proof (rule exI[of _ \<tau>2], intro conjI)
    show "\<forall> x \<in> vars_term (Fun f ts). \<tau>2 x \<in> ta_res TA (\<sigma> x)"
    proof
      fix x
      assume "x \<in> vars_term (Fun f ts)"
      then obtain t where "t \<in> set ts" and "x \<in> vars_term t" by auto
      then obtain i where i: "i < length ts" and x: "x \<in> vars_term (ts ! i)" unfolding set_conv_nth by auto
      from ind[OF i] \<tau>2[OF i] x show "\<tau>2 x \<in> ta_res TA (\<sigma> x)" by auto
    qed
  next
    { 
      fix i
      assume i: "i < length ts"
      have id: "map_vars \<tau>2 (ts ! i) = map_vars (\<tau> i) (ts ! i)" 
        by (rule map_vars_vars_term, insert \<tau>2[OF i], auto)
      from ind[OF i, THEN conjunct1] 
      have "qs ! i \<in> ta_res TA (map_vars \<tau>2 (ts ! i))" unfolding id .
    } note main = this
    have "q \<in> ta_res TA (Fun f (map (map_vars \<tau>2) ts))"
      by (unfold ta_res.simps, rule, rule exI[of _ q], rule exI[of _ q'], rule exI[of _ qs], intro conjI, simp, rule rule, rule q, simp add: len, simp add: main)
    thus "q \<in> ta_res TA (map_vars \<tau>2 (Fun f ts))" by auto
  qed
qed

(* decomposing substitutions: deterministic case *)
lemma ta_res_subst_det: fixes \<sigma> TA
  defines "\<tau> \<equiv> (\<lambda> x. SOME q. q \<in> ta_res TA (\<sigma> x))"
  assumes det: "ta_det TA"
  and q: "q \<in> ta_res TA (t \<cdot> \<sigma>)"
  shows "q \<in> ta_res TA (map_vars \<tau> t) \<and> (\<forall> x \<in> vars_term t. \<tau> x \<in> ta_res TA (\<sigma> x))"
using q
proof (induct t arbitrary: q)
  case (Var x)
  hence "q \<in> ta_res TA (\<sigma> x)" by auto
  hence tau: "\<tau> x \<in> ta_res TA (\<sigma> x)" unfolding \<tau>_def by (rule someI)
  from det this Var have "q = \<tau> x" by auto
  with tau show ?case by auto
next
  case (Fun f ts)
  from Fun(2)
  obtain q' qs where rule: "(f qs \<rightarrow> q') \<in> ta_rules TA"
    and len: "length qs = length ts"
    and rec: "\<And>i. i < length ts \<Longrightarrow> qs ! i \<in> ta_res TA (ts ! i \<cdot> \<sigma>)"
    and q: "(q',q) \<in> (ta_eps TA)^*"
    by auto
  {
    fix i
    assume i: "i < length ts" 
    from i have mem: "ts ! i \<in> set ts" by simp
    note Fun(1)[OF mem rec[OF i]]
  }
  hence
    ind: "\<And> i. i < length ts \<Longrightarrow> qs ! i \<in> ta_res TA (map_vars \<tau> (ts ! i)) \<and> 
             (\<forall> x \<in> vars_term (ts ! i). \<tau> x \<in> ta_res TA (\<sigma> x))"
    by auto
  show ?case
  proof (intro conjI)
    show "\<forall> x \<in> vars_term (Fun f ts). \<tau> x \<in> ta_res TA (\<sigma> x)"
    proof
      fix x
      assume "x \<in> vars_term (Fun f ts)"
      then obtain t where "t \<in> set ts" and "x \<in> vars_term t" by auto
      then obtain i where i: "i < length ts" and x: "x \<in> vars_term (ts ! i)" unfolding set_conv_nth by auto
      from ind[OF i] x show "\<tau> x \<in> ta_res TA (\<sigma> x)" by auto
    qed
  next
    from ind have main: "\<And> i. i < length ts \<Longrightarrow> qs ! i \<in> ta_res TA (map_vars \<tau> (ts ! i))" ..
    have "q \<in> ta_res TA (Fun f (map (map_vars \<tau>) ts))"
      by (unfold ta_res.simps, rule, rule exI[of _ q], rule exI[of _ q'], rule exI[of _ qs], 
        intro conjI, simp, rule rule, rule q, simp add: len, simp add: main)
    thus "q \<in> ta_res TA (map_vars \<tau> (Fun f ts))" by auto
  qed
qed

lemma ta_res_subst_linear_det: 
  assumes ld: "linear_term t \<or> ta_det TA"
  and q: "q \<in> ta_res TA (t \<cdot> \<sigma>)"
  shows "\<exists> \<tau>. q \<in> ta_res TA (map_vars \<tau> t) \<and> (\<forall> x \<in> vars_term t. \<tau> x \<in> ta_res TA (\<sigma> x))"
  using ld
proof 
  assume "linear_term t"
  from ta_res_subst_linear[OF this q] show ?thesis .
next
  assume "ta_det TA"
  from ta_res_subst_det[OF this q] show ?thesis by - (rule exI)
qed

(* reachable states for some term within a context *)
lemma ta_res_ctxt: assumes p: "p \<in> ta_res TA t"
  shows "q \<in> ta_res TA (C \<langle> Var p \<rangle>) \<Longrightarrow> q \<in> ta_res TA (C \<langle> t \<rangle>)"
proof (induct C arbitrary: q)
  case Hole
  thus ?case using p ta_res_eps by auto
next
  case (More f bef C aft)
  let ?n = "length bef"
  let ?m = "Suc (?n + length aft)"
  from More(2)
  obtain q' qs where 
    rule: "(f qs \<rightarrow> q') \<in> ta_rules TA" and
    eps: "(q', q) \<in> (ta_eps TA)\<^sup>*" and
    len: "length qs = ?m" and
    rec: "\<And> i. i < ?m \<Longrightarrow> qs ! i \<in> (map (ta_res TA) bef @ ta_res TA C\<langle>Var p\<rangle> # map (ta_res TA) aft) ! i" 
    by auto
  from rec[of ?n] len have "qs ! ?n \<in> ta_res TA (C \<langle> Var p \<rangle>)" by (simp add: nth_append)
  from More(1)[OF this] have IH: "qs ! ?n \<in> ta_res TA C\<langle>t\<rangle>" .
  show ?case 
  proof (simp, intro exI conjI, rule rule, rule eps, rule len, clarify)
    fix i
    assume "i < ?m"
    from rec[OF this] IH
    show "qs ! i \<in> (map (ta_res TA) bef @ ta_res TA C\<langle>t\<rangle> # map (ta_res TA) aft) ! i"
      by (cases "i = ?n", auto simp: nth_append)
  qed
qed

(* decomposing contexts *)
lemma ta_res_ctxt_decompose: assumes "q \<in> ta_res TA (C \<langle> t \<rangle>)"
  shows "\<exists> p . p \<in> ta_res TA t \<and> q \<in> ta_res TA (C \<langle> Var p \<rangle>)"
  using assms
proof (induct C arbitrary: q)
  case (More f bef C aft q)
  let ?n = "length bef"
  let ?m = "Suc (?n + length aft)"
  from More(2)[simplified]
  obtain q' qs where 
    rule: "(f qs \<rightarrow> q') \<in> ta_rules TA" and
    eps: "(q', q) \<in> (ta_eps TA)\<^sup>*" and
    len: "length qs = ?m" and
    rec: "\<And> i. i < ?m \<Longrightarrow> qs ! i \<in> (map (ta_res TA) bef @ ta_res TA C\<langle>t\<rangle> # map (ta_res TA) aft) ! i" 
    by auto
  from rec[of ?n] len have "qs ! ?n \<in> ta_res TA (C \<langle>t\<rangle>)" by (simp add: nth_append)
  from More(1)[OF this] obtain p where p: "p \<in> ta_res TA t" and IH: "qs ! ?n \<in> ta_res TA (C \<langle>Var p\<rangle>)" by auto
  show ?case
  proof (intro exI conjI, rule p, simp, intro exI conjI, rule rule, rule eps, rule len, clarify)
    fix i
    assume "i < ?m"
    from rec[OF this] IH
    show "qs ! i \<in> (map (ta_res TA) bef @ ta_res TA C\<langle>Var p\<rangle> # map (ta_res TA) aft) ! i"
      by (cases "i = ?n", auto simp: nth_append)
  qed
qed auto

(* a resulting state is always some rhs of a rule (or epsilon transition) *)
lemma ta_rhs_states_res: assumes "is_Fun t" 
  shows "ta_res TA t \<subseteq> ta_rhs_states TA"
proof
  fix q 
  assume q: "q \<in> ta_res TA t"
  from `is_Fun t` obtain f ts where t: "t = Fun f ts" by (cases t, auto)
  from q[unfolded t] obtain q' qs where "(f qs \<rightarrow> q') \<in> ta_rules TA" 
    and q: "(q',q) \<in> (ta_eps TA)^*" by auto
  thus "q \<in> ta_rhs_states TA" unfolding ta_rhs_states_def by force
qed    

(* only terms can be reduced to a state, if all symbols appear in the automaton *)
lemma ta_syms_res: "q \<in> ta_res TA t \<Longrightarrow> funas_term t \<subseteq> ta_syms TA"
proof (induct t arbitrary: q)
  case (Var x)
  show ?case by simp
next
  case (Fun f ts)
  from Fun(2)
  obtain q' qs where rule: "(f qs \<rightarrow> q') \<in> ta_rules TA"
    and len: "length qs = length ts"
    and rec: "\<forall> i < length ts. qs ! i \<in> ta_res TA (ts ! i)"
    by auto
  {
    fix t 
    assume mem: "t \<in> set ts"
    then obtain i where i: "i < length ts"
      and t: "t = ts ! i" unfolding set_conv_nth by auto
    from Fun(1)[OF mem[unfolded t] rec[THEN spec, THEN mp[OF _ i]]]
    have "funas_term t \<subseteq> ta_syms TA" unfolding t by auto
  } note rec = this
  from rule len have f: "(f,length ts) \<in> ta_syms TA" unfolding ta_syms_def
    by force
  show ?case using f rec by auto
qed

(* only terms can be accepted, if all symbols appear in the automaton *)
lemma ta_syms_lang: assumes "t \<in> ta_lang TA" shows "funas_term t \<subseteq> ta_syms TA"
proof -
  from assms
  obtain q where "q \<in> ta_res TA t" unfolding ta_lang_def
    by auto
  from ta_syms_res[OF this] show ?thesis  .
qed

subsection {* Theorem 2: state-coherence and state-compatibility implies closed under rewriting *}

(* state-compatibility for a rule,
   the restrictions on vars_term and \<tau> are made to ensure that we will only
   have to test finitely many \<tau> *)
definition rule_state_compatible :: "('q,'f)ta \<Rightarrow> 'q rel \<Rightarrow> ('f,'v)rule \<Rightarrow> bool"
where "rule_state_compatible TA rel \<equiv> \<lambda> (l,r). (\<forall> \<tau>. (\<tau> ` vars_term l \<subseteq> ta_rhs_states TA) \<longrightarrow> 
  ta_res TA (map_vars \<tau> l) \<subseteq> rel^-1 `` ta_res TA (map_vars \<tau> r))"

(* corresponds to Definition 3 in paper *)
definition state_coherent :: "('q,'f)ta \<Rightarrow> 'q rel \<Rightarrow> bool" where
  "state_coherent TA rel \<equiv> rel `` ta_final TA \<subseteq> ta_final TA \<and> (\<forall> f qs q i qi. 
  (f qs \<rightarrow> q) \<in> ta_rules TA \<longrightarrow> i < length qs \<longrightarrow> (qs ! i, qi) \<in> rel \<longrightarrow> 
  (\<exists> q'. (f (qs[ i := qi]) \<rightarrow> q') \<in> ta_rules TA \<and> (q,q') \<in> rel))"

(* corresponds to Definition 3 in paper *)
definition state_compatible :: "('q,'f)ta \<Rightarrow> 'q rel \<Rightarrow> ('f,'v)trs \<Rightarrow> bool"
where "state_compatible TA rel R \<equiv> \<forall> l r. (l,r) \<in> R \<and> funas_term l \<subseteq> ta_syms TA 
  \<longrightarrow> rule_state_compatible TA rel (l,r)" 

lemma state_coherent_Id: "state_coherent TA Id" unfolding state_coherent_def by auto

(* major lemma, corresponds to proof of Theorem 2 *)
lemma state_compatible_res: assumes 
      comp: "state_compatible TA rel R"
  and coh: "state_coherent TA rel"
  and ll_or_det: "(left_linear_trs R \<and> (rel = Id \<or> ta_eps TA = {})) \<or> ta_det TA"
  and wf_trs: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and step: "(s,t) \<in> rstep R"
  and ground: "ground s"
  shows "ta_res TA s \<subseteq> rel^-1 `` ta_res TA t"
  using step ground
proof (induct)
  case (IH C \<sigma> l r)
  hence ground: "ground (l \<cdot> \<sigma>)" by simp
  show ?case
  proof (induct C)
    case (More f bef C aft)
    let ?args = "\<lambda> l. bef @ C\<langle>l \<cdot> \<sigma>\<rangle> # aft"
    let ?l = "Fun f (?args l)"
    let ?r = "Fun f (?args r)"
    let ?n = "Suc (length bef + length aft)"
    let ?m = "length bef"
    show ?case
    proof
      fix q
      assume "q \<in> ta_res TA (More f bef C aft)\<langle>l\<cdot>\<sigma>\<rangle>"
      hence "q \<in> ta_res TA ?l" by simp
      from this[unfolded ta_res.simps] obtain q' qs where 
        rule: "(f qs \<rightarrow> q') \<in> ta_rules TA"
        and len: "length qs = ?n"
        and rec: "\<forall>i < ?n. qs ! i \<in> map (ta_res TA) (?args l) ! i"
        and q: "(q',q) \<in> (ta_eps TA)^*"
        by auto
      have len2: "\<And>t. ?m < length (?args t)" by auto
      from rec[THEN spec, of ?m, unfolded nth_map[OF len2]]
      have "qs ! ?m \<in> ta_res TA (C\<langle>l \<cdot> \<sigma>\<rangle>)"  by auto
      with More have m: "qs ! ?m \<in> rel^-1 `` ta_res TA (C\<langle>r\<cdot>\<sigma>\<rangle>)" by auto
      then obtain qm where rel: "(qs ! ?m,qm) \<in> rel" and m: "qm \<in> ta_res TA (C\<langle>r\<cdot>\<sigma>\<rangle>)" by auto
      let ?qs = "qs[ ?m := qm]"
      { 
        fix i
        assume i: "i < ?n"
        have "?qs ! i \<in> map (ta_res TA) (?args r) ! i" 
        proof (cases "i = ?m")
          case True
          hence id: "?qs ! i = qm" using len by auto
          show ?thesis unfolding id using m unfolding True 
            by (simp add: nth_append)
        next
          case False
          hence id: "?qs ! i = qs ! i" by auto
          note rec = rec[THEN spec, THEN mp, OF i]
          show ?thesis 
          proof (cases "i < ?m")
            case True
            show ?thesis using rec unfolding id
              by (simp add: nth_append True)
          next
            case False
            with `i \<noteq> ?m` have "i - ?m > 0" by auto
            then obtain j where im: "i - ?m = Suc j" by (cases "i - ?m", auto)
            show ?thesis using rec unfolding id
              by (simp add: nth_append False im)
          qed
        qed
      }
      hence rec: "\<forall>i < ?n. ?qs ! i \<in> map (ta_res TA) (?args r) ! i"
        by simp 
      from len have "?m < length qs" by auto
      from coh[unfolded state_coherent_def, THEN conjunct2, rule_format, OF rule this rel]
        obtain q'' where rule: "(f ?qs \<rightarrow> q'') \<in> ta_rules TA" and rel: "(q', q'') \<in> rel" by auto
      have q'': "q'' \<in> ta_res TA ?r" 
        unfolding ta_res.simps
        by (rule, intro exI conjI, rule refl, rule rule, insert len rec, auto)
      from ll_or_det have "rel = Id \<or> ta_eps TA = {}" unfolding ta_det_def by auto      
      thus "q \<in> rel^-1 `` ta_res TA (More f bef C aft)\<langle>r\<cdot>\<sigma>\<rangle>" 
      proof 
        assume "ta_eps TA = {}"
        with q q'' rel show ?thesis by auto
      next
        assume "rel = Id"
        with rel have id: "q' = q''" by simp
        from ta_res_eps[OF q'' q[unfolded id]] `rel = Id` show ?thesis by auto
      qed
    qed
  next
    case Hole
    show ?case
    proof (unfold ctxt_apply.simps, rule)
      fix q
      assume q: "q \<in> ta_res TA (l \<cdot> \<sigma>)"
      from ll_or_det IH(1) have lin: "linear_term l \<or> ta_det TA" unfolding left_linear_trs_def by auto
      from ta_res_subst_linear_det[OF this q] obtain \<tau> where
        q: "q \<in> ta_res TA (map_vars \<tau> l)" and 
        vars: "\<forall> x \<in> vars_term l. \<tau> x \<in> ta_res TA (\<sigma> x)" by auto
      from ta_syms_res[OF q]
      have "funas_term l \<subseteq> ta_syms TA" by auto
      with comp[unfolded state_compatible_def] IH(1) have
        comp: "rule_state_compatible TA rel (l,r)" by auto
      have "\<tau> ` vars_term l \<subseteq> ta_rhs_states TA" 
      proof 
        fix q
        assume "q \<in> \<tau> ` vars_term l"
        then obtain x where x: "x \<in> vars_term l" and q: "q = \<tau> x" by auto
        with vars have q: "q \<in> ta_res TA (\<sigma> x)" by auto
        from ground x have "ground (\<sigma> x)" by auto
        hence "is_Fun (\<sigma> x)" by auto
        from ta_rhs_states_res[OF this] q show "q \<in> ta_rhs_states TA" by auto
      qed
      with comp[unfolded rule_state_compatible_def] q
      have q: "q \<in> rel^-1 `` ta_res TA (map_vars \<tau> r)" by auto
      then obtain q' where qq': "(q,q') \<in> rel" and q': "q' \<in> ta_res TA (map_vars \<tau> r)" by auto
      from ta_res_subst[OF _ q'] wf_trs[OF IH(1)] vars qq'
      show "q \<in> rel^-1 `` ta_res TA (r \<cdot> \<sigma>)" by force
    qed
  qed
qed

(* Theorem 2 in paper: soundness *)
lemma state_compatible_lang: assumes 
      comp: "state_compatible TA rel R"
  and coh: "state_coherent TA rel"
  and ll_or_det: "(left_linear_trs R \<and> (rel = Id \<or> ta_eps TA = {})) \<or> ta_det TA"
  and wf_trs: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and step: "(s,t) \<in> rstep R"
  and s: "s \<in> ta_lang TA"
  shows "t \<in> ta_lang TA"
proof -
  from s[unfolded ta_lang_def] obtain q where
    g: "ground s" and fin: "q \<in> ta_final TA" and res: "q \<in> ta_res TA s" by auto
  from rstep_ground[OF wf_trs g step] have gt: "ground t" .
  from state_compatible_res[OF comp coh ll_or_det wf_trs step g] res obtain q' 
    where qq': "(q,q') \<in> rel" and res: "q' \<in> ta_res TA t" by auto
  from coh[unfolded state_coherent_def] fin qq' have fin: "q' \<in> ta_final TA" by auto
  from fin gt res show "t \<in> ta_lang TA" unfolding ta_lang_def by auto
qed


(* Theorem 2 applied to have reachability *)
lemma state_compatible: assumes 
      comp: "state_compatible TA rel R"
  and coh: "state_coherent TA rel"
  and ll_or_det: "(left_linear_trs R \<and> (rel = Id \<or> ta_eps TA = {})) \<or> ta_det TA"
  and wf_trs: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and L: "L \<subseteq> ta_lang TA"
  shows "{t | s t. s \<in> L \<and> (s,t) \<in> (rstep R)^*} \<subseteq> ta_lang TA"
proof -
  {
    fix s t
    assume "(s,t) \<in> (rstep R)^*" and "s \<in> L" 
    hence "t \<in> ta_lang TA"
    proof (induct)
      case base
      thus ?case using L by auto
    next
      case (step t u)
      hence t: "t \<in> ta_lang TA" by simp
      show ?case
        by (rule state_compatible_lang[OF comp coh ll_or_det wf_trs step(2) t])
    qed
  }
  thus ?thesis by auto
qed

subsection {* Proposition 1: there always is a trim automaton *}

definition "ta_subset TA TA' \<equiv> ta_final TA \<subseteq> ta_final TA' \<and> ta_rules TA \<subseteq> ta_rules TA' \<and> ta_eps TA \<subseteq> ta_eps TA'"

lemma ta_res_mono: assumes "ta_subset TA TA'"
  shows "ta_res TA t \<subseteq> ta_res TA' t"
proof -
  from assms[unfolded ta_subset_def] have eps: "ta_eps TA \<subseteq> ta_eps TA'" and rules: "ta_rules TA \<subseteq> ta_rules TA'" by auto
  note eps = rtrancl_mono[OF eps]
  show ?thesis
  proof (induct t)
    case (Fun f ts)
    show ?case 
    proof
      fix q 
      assume "q \<in> ta_res TA (Fun f ts)"
      from this[simplified] obtain qa qs where
        rule: "(f qs \<rightarrow> qa) \<in> ta_rules TA \<and> (qa, q) \<in> (ta_eps TA)\<^sup>*"  
        and len: "length qs = length ts" 
        and steps: "\<And> i. i<length ts \<Longrightarrow> qs ! i \<in> ta_res TA (ts ! i)" by auto
      with eps rules have 
        rule: "(f qs \<rightarrow> qa) \<in> ta_rules TA' \<and> (qa, q) \<in> (ta_eps TA')\<^sup>*" by auto
      {
        fix i
        assume "i < length ts"
        with Fun[unfolded set_conv_nth, of "ts ! i"] steps[OF this]
        have "qs ! i \<in> ta_res TA' (ts ! i)" by auto
      }
      with rule len show "q \<in> ta_res TA' (Fun f ts)" by auto
    qed
  qed (insert eps, auto)
qed

lemma ta_lang_mono: assumes "ta_subset TA TA'"
  shows "ta_lang TA \<subseteq> ta_lang TA'"
  using ta_res_mono[OF assms] assms[unfolded ta_subset_def] unfolding ta_lang_def 
  by blast

(* the restriction of an automata to a given set of states *)
definition ta_restrict where 
  "ta_restrict TA Q = \<lparr> ta_final = ta_final TA \<inter> Q, 
    ta_rules = { r. r \<in> ta_rules TA \<and> r_states r \<subseteq> Q },
    ta_eps = ta_eps TA \<inter> (Q \<times> Q)
    \<rparr>"

lemma ta_restrict_subset: "ta_subset (ta_restrict TA Q) TA"
  unfolding ta_subset_def ta_restrict_def Let_def by auto

lemma ta_restrict_states_Q: "ta_states (ta_restrict TA Q) \<subseteq> Q"
  unfolding ta_states_def ta_restrict_def Let_def by auto

lemma ta_restrict_states: "ta_states (ta_restrict TA Q) \<subseteq> ta_states TA"
  unfolding ta_states_def ta_restrict_def Let_def by auto


lemma ta_subset_refl[simp]: "ta_subset TA TA" 
  unfolding ta_subset_def by auto

lemma ta_subset_trans: "ta_subset TA TA' \<Longrightarrow> ta_subset TA' TA'' \<Longrightarrow> ta_subset TA TA''"
  unfolding ta_subset_def by auto

lemma ta_restrict_states_eq_imp_eq: assumes eq: "ta_states (ta_restrict TA Q) = ta_states TA"
  shows "ta_restrict TA Q = TA"
  by (cases TA, insert assms, unfold ta_restrict_def ta_states_def, auto)

lemma ta_subset_det: "ta_subset TA TA' \<Longrightarrow> ta_det TA' \<Longrightarrow> ta_det TA"
  unfolding ta_det_def ta_subset_def by blast

definition ta_reachable :: "('q,'f)ta \<Rightarrow> 'q \<Rightarrow> bool" where
  "ta_reachable TA q \<equiv> \<exists> t. ground t \<and> q \<in> ta_res TA t"

lemma ta_reachableE: assumes "ta_reachable TA q"
  obtains t where "ground t" "q \<in> ta_res TA t"
  using assms[unfolded ta_reachable_def] by auto

definition ta_productive :: "('q,'f)ta \<Rightarrow> 'q \<Rightarrow> bool" where
  "ta_productive TA q \<equiv> \<exists> q' q'' C. (q,q') \<in> (ta_eps TA)^* \<and> q'' \<in> ta_res TA (C\<langle>Var q'\<rangle>) \<and> q'' \<in> ta_final TA \<and> ground_ctxt C"

lemma ta_productiveE: assumes "ta_productive TA q"
  obtains q' q'' C where "(q,q') \<in> (ta_eps TA)^*" "q'' \<in> ta_res TA (C\<langle>Var q'\<rangle>)" "q'' \<in> ta_final TA" "ground_ctxt C"
  using assms[unfolded ta_productive_def] by auto

lemma ta_productiveI: 
  assumes "(q,q') \<in> (ta_eps TA)^*" "q'' \<in> ta_res TA (C\<langle>Var q'\<rangle>)" "q'' \<in> ta_final TA" "ground_ctxt C"
  shows "ta_productive TA q"
  using assms unfolding ta_productive_def by auto

lemma ta_productiveI': 
  assumes "q' \<in> ta_res TA (C\<langle>Var q\<rangle>)" "q' \<in> ta_final TA" "ground_ctxt C"
  shows "ta_productive TA q"
    by (rule ta_productiveI[OF _ assms], auto)

(* Definition 2 in paper *)
definition ta_trim :: "('q,'f)ta \<Rightarrow> bool" where 
  "ta_trim TA \<equiv> \<forall> q \<in> ta_states TA. ta_reachable TA q \<and> ta_productive TA q"


abbreviation ta_only_reach :: "('q,'f)ta \<Rightarrow> ('q,'f)ta" where
  "ta_only_reach TA \<equiv> ta_restrict TA {q. ta_reachable TA q}"

lemma ta_reachable_eps: assumes Q: "Q = {q. ta_reachable TA q}"
  and q: "q \<in> Q" shows "(q,q') \<in> (ta_eps TA)^* \<Longrightarrow> q' \<in> Q \<and> (q,q') \<in> (ta_eps TA \<inter> Q \<times> Q)^*"
proof (induct rule: rtrancl_induct)
  case (step q' q'')
  from q[unfolded Q] have "ta_reachable TA q" by auto
  from ta_reachableE[OF this] obtain t where gt: "ground t" and qt: "q \<in> ta_res TA t" by auto
  let ?rel = "(ta_eps TA \<inter> Q \<times> Q)"
  from step(1-2) have "(q,q'') \<in> (ta_eps TA)^*" by auto
  from ta_res_eps[OF qt this] gt have "ta_reachable TA q''" unfolding ta_reachable_def by auto
  with step(2) have q'': "q'' \<in> Q" unfolding Q ta_states_def by auto
  with step(2-3) have "(q,q') \<in> ?rel^*" "(q',q'') \<in> ?rel" by auto
  hence "(q,q'') \<in> ?rel^*" by (rule rtrancl_into_rtrancl)
  with q'' show ?case by simp
qed (insert q, auto)

(* major lemma to show that one can restrict to reachable states *)
lemma ta_res_only_reach: assumes Q: "Q = {q. ta_reachable TA q}"
  shows "vars_term t \<subseteq> Q \<Longrightarrow> q \<in> ta_res TA t \<Longrightarrow> q \<in> Q \<and> q \<in> ta_res (ta_only_reach TA) t"
proof (induct t arbitrary: q)
  case (Var q q')
  from Var(2) have qq': "(q,q') \<in> (ta_eps TA)^*" by simp
  from Var(1) have q: "q \<in> Q" by simp
  from ta_reachable_eps[OF Q q qq'] show ?case unfolding ta_restrict_def Let_def Q by simp
next
  case (Fun f ts)
  let ?TA = "ta_only_reach TA"
  from Fun(3)[simplified] obtain qa qs where
        rule: "(f qs \<rightarrow> qa) \<in> ta_rules TA" and eps: "(qa, q) \<in> (ta_eps TA)\<^sup>*"  
        and len: "length qs = length ts" 
        and steps: "\<And> i. i<length ts \<Longrightarrow> qs ! i \<in> ta_res TA (ts ! i)" by auto
  {
    fix i
    assume i: "i < length ts"
    hence mem: "ts ! i \<in> set ts" by auto
    with Fun(2) have "vars_term (ts ! i) \<subseteq> Q" by auto
    from Fun(1)[OF mem this steps[OF i]]
    have qi: "qs ! i \<in> Q" and qti: "qs ! i \<in> ta_res ?TA (ts ! i)" by auto
    from qi[unfolded Q] have "ta_reachable TA (qs ! i)" by auto
    from ta_reachableE[OF this] have "\<exists> gi. ground gi \<and> qs ! i \<in> ta_res TA gi" by auto
    note this qi qti
  } note IH = this
  from IH(1) have "\<forall> i. \<exists> g. i < length ts \<longrightarrow> ground g \<and> qs ! i \<in> ta_res TA g" by auto
  from choice[OF this] obtain gs where gs: "\<And> i. i < length ts \<Longrightarrow> ground (gs i) \<and> qs ! i \<in> ta_res TA (gs i)" by auto
  let ?g = "Fun f (map gs [0 ..< length ts])"
  have g: "ground ?g" using gs by auto
  have "qa \<in> ta_res TA ?g"
    by (simp, intro exI conjI, rule rule, insert len gs, auto)
  with g have reach: "ta_reachable TA qa" unfolding ta_reachable_def by blast 
  with Q have qa: "qa \<in> Q" by auto
  with IH(2) len have "r_states (f qs \<rightarrow> qa) \<subseteq> Q" unfolding r_states_def by (auto simp: set_conv_nth)
  with rule have rule: "(f qs \<rightarrow> qa) \<in> ta_rules ?TA" unfolding ta_restrict_def Let_def Q by auto
  from ta_reachable_eps[OF Q qa eps] have q: "q \<in> Q" and eps: "(qa, q) \<in> (ta_eps ?TA)\<^sup>*" 
    unfolding ta_restrict_def Let_def Q by auto
  have "q \<in> ta_res ?TA (Fun f ts)"
    by (simp, intro exI conjI, rule rule, rule eps, insert len IH(3), auto)
  with q show ?case by auto
qed  

(* It is sound to restrict to reachable states. *)
lemma ta_only_reach_lang: "ta_lang (ta_only_reach TA) = ta_lang TA" (is "ta_lang ?TA = _")
proof
  show "ta_lang ?TA \<subseteq> ta_lang TA" by (rule ta_lang_mono[OF ta_restrict_subset])
  show "ta_lang TA \<subseteq> ta_lang ?TA" 
  proof
    fix t
    assume "t \<in> ta_lang TA"
    from this[unfolded ta_lang_def] obtain q where gt: "ground t" and f: "q \<in> ta_final TA" 
    and q: "q \<in> ta_res TA t" by auto
    have "q \<in> {q. ta_reachable TA q} \<and> q \<in> ta_res ?TA t"
      by (rule ta_res_only_reach[OF refl _ q], insert gt, auto simp: ground_vars_term_empty)
    with f gt show "t \<in> ta_lang ?TA" unfolding ta_restrict_def ta_lang_def Let_def by auto
  qed
qed

abbreviation ta_only_prod :: "('q,'f)ta \<Rightarrow> ('q,'f)ta" where
  "ta_only_prod TA \<equiv> ta_restrict TA {q. ta_productive TA q}"

(* Major lemma to show that it is sound to restrict to productive states. *)
lemma ta_res_only_prod: assumes Q: "Q = {q. ta_productive TA q}"
  shows "q \<in> ta_res TA t \<Longrightarrow> ta_productive TA q \<Longrightarrow> ground t \<Longrightarrow> q \<in> ta_res (ta_only_prod TA) t"
proof (induct t arbitrary: q)
  case (Fun f ts)
  let ?TA = "ta_only_prod TA"
  from Fun(2) obtain qa qs where
        rule: "(f qs \<rightarrow> qa) \<in> ta_rules TA" and eps: "(qa, q) \<in> (ta_eps TA)\<^sup>*"  
        and len: "length qs = length ts" 
        and steps: "\<And> i. i<length ts \<Longrightarrow> qs ! i \<in> ta_res TA (ts ! i)" by auto
  from Fun(3) have q: "ta_productive TA q" .
  from ta_productiveE[OF q] obtain q' q'' C where qq': "(q,q') \<in> (ta_eps TA)^*" and qq'': "q'' \<in> ta_res TA (C \<langle>Var q'\<rangle>)" and q'': "q'' \<in> ta_final TA"
    and gC: "ground_ctxt C" by auto
  {
    fix i
    assume i: "i < length ts"
    hence "ts ! i \<in> set ts" by auto
    with Fun(4) have "ground (ts ! i)" by auto
  } note ground = this
  from eps qq' have qaq': "(qa,q') \<in> (ta_eps TA)^*" by auto
  {
    fix i
    assume i: "i < length ts"
    hence [simp]: "Suc (i + (length ts - Suc i)) = length ts" "min (length ts) i = i" by arith+
    let ?D = "C \<circ>c More f (take i ts) \<box> (drop (Suc i) ts)"
    have gD: "ground_ctxt ?D" using gC ground i 
      by (auto simp: set_conv_nth)
    have id: "(C \<circ>c More f (take i ts) \<box> (drop (Suc i) ts))\<langle>Var (qs ! i)\<rangle> = C\<langle>Fun f (take i ts @ Var (qs ! i) # drop (Suc i) ts)\<rangle>" by simp 
    have [simp]: "qs ! i \<in> ta_res TA (Var (qs ! i))" by auto
    note ta_res.simps(1)[simp del]
    have qsi: "ta_productive TA (qs ! i)"
    proof (rule ta_productiveI'[OF _ q'' gD], unfold id, rule ta_res_ctxt[OF _ qq''], simp, intro exI conjI,
      rule rule, rule qaq', simp add: len, intro allI impI)
      fix j
      assume j: "j < length ts"
      let ?l = "map (ta_res TA) (take i ts) @ ta_res TA (Var (qs ! i)) # map (ta_res TA) (drop (Suc i) ts)"
      show "qs ! j \<in> ?l ! j"
      proof (cases "j = i")
        case False
        with j i have "?l ! j = ta_res TA (ts ! j)" by (auto simp: nth_append)
        with steps[OF j] show ?thesis by auto
      next
        case True
        with i have "?l ! j = ta_res TA (Var (qs ! i))" by (auto simp: nth_append)
        thus ?thesis using True by simp
      qed
    qed
    from Fun(1)[OF _ steps[OF i] qsi ground[OF i]] i have "qs ! i \<in> ta_res ?TA (ts ! i)" by auto
    note this qsi
  } note IH = this
  have qa: "ta_productive TA qa"
    by (rule ta_productiveI[OF qaq' qq'' q'' gC])
  with IH(2) len have "r_states (f qs \<rightarrow> qa) \<subseteq> Q" unfolding Q by (auto simp: r_states_def set_conv_nth)
  with rule have rule: "(f qs \<rightarrow> qa) \<in> ta_rules ?TA" unfolding ta_restrict_def Q Let_def by auto
  from eps have "(qa,q) \<in> (ta_eps ?TA)^* \<and> qa \<in> Q"
  proof (induct rule: converse_rtrancl_induct)
    case base
    show ?case using q unfolding Q by auto
  next
    case (step q1 q2)
    have "ta_productive TA q1"
      by (rule ta_productiveI[OF _ qq'' q'' gC], insert step(1-2) qq', auto)
    with step(1,3) have "(q1,q2) \<in> ta_eps ?TA" "q1 \<in> Q" unfolding Q ta_restrict_def Let_def by auto
    with step(3) show ?case by auto
  qed
  hence eps: "(qa,q) \<in> (ta_eps ?TA)^*" ..
  show ?case
    by (simp, intro exI conjI, rule rule, rule eps, insert len IH(1), auto)
qed auto

lemma ta_productive_final: assumes q: "q \<in> ta_final TA"
  shows "ta_productive TA q"
  by (rule ta_productiveI'[OF _ q, of \<box>], auto)

(* It is sound to restrict to productive states. *)
lemma ta_only_prod_lang: "ta_lang (ta_only_prod TA) = ta_lang TA" (is "ta_lang ?TA = _")
proof
  show "ta_lang ?TA \<subseteq> ta_lang TA" by (rule ta_lang_mono[OF ta_restrict_subset])
  show "ta_lang TA \<subseteq> ta_lang ?TA" 
  proof
    fix t
    assume "t \<in> ta_lang TA"
    from this[unfolded ta_lang_def] obtain q where gt: "ground t" and f: "q \<in> ta_final TA" 
      and q: "q \<in> ta_res TA t" by auto
    from ta_res_only_prod[OF refl q ta_productive_final[OF f] gt]
    have q: "q \<in> ta_res ?TA t" .
    from f ta_productive_final[of _ TA] have "q \<in> ta_final ?TA"
      unfolding ta_restrict_def Let_def by auto
    with q gt show "t \<in> ta_lang ?TA" unfolding ta_lang_def by auto
  qed
qed

(* turn a finite automaton into a trim one, by removing
   unreachable and non-productive states as long as possible *)
function trim_ta :: "('q,'f)ta \<Rightarrow> ('q,'f)ta" where
  "trim_ta TA = (if finite (ta_states TA) 
    then (let TA' = ta_only_prod (ta_only_reach TA) 
      in if ta_states TA = ta_states TA' then TA else trim_ta TA') else TA)"
  by pat_completeness auto

lemma ta_prod_reach_states: "ta_states (ta_only_prod (ta_only_reach TA)) \<subseteq> ta_states TA"
  by (rule subset_trans, (rule ta_restrict_states)+)

termination
proof
  fix TA TA' :: "('q,'f)ta"
  let ?R = "ta_only_reach TA"
  let ?PR = "ta_only_prod ?R"
  assume fin: "finite (ta_states TA)" 
    and ta': "TA' = ?PR" 
    and neq: "ta_states TA \<noteq> ta_states TA'"
  have "ta_states TA' \<subseteq> ta_states TA" 
    unfolding ta' 
    by (rule ta_prod_reach_states)
  with neq
  have "ta_states TA' \<subset> ta_states TA" by blast
  with fin have "(ta_states TA', ta_states TA) \<in> finite_psubset" by auto
  thus "(TA', TA) \<in> inv_image finite_psubset ta_states" by auto
qed auto

lemma trim_ta_lang: "ta_lang (trim_ta TA) = ta_lang TA"
proof (induct rule: trim_ta.induct)
  case (1 TA)
  note trim_ta.simps[simp del]
  note trim_ta.simps[of TA, simp]
  show ?case
  proof (cases "finite (ta_states TA)")
    case False
    thus ?thesis by auto
  next
    case True
    note IH = 1[OF this refl]
    note ta_only_reach_lang[simp] ta_only_prod_lang[simp]
    let ?R = "ta_only_reach TA"
    let ?PR = "ta_only_prod ?R"
    show ?thesis
    proof (cases "ta_states TA = ta_states ?PR")
      case True
      thus ?thesis by auto
    next
      case False
      note IH = IH[OF this]
      with False True show ?thesis by simp
    qed
  qed
qed


lemma ta_prod_reach_subset: "ta_subset (ta_only_prod (ta_only_reach TA)) TA"
  by (rule ta_subset_trans, (rule ta_restrict_subset)+)

lemma trim_ta_subset: "ta_subset (trim_ta TA) TA"
proof (induct rule: trim_ta.induct)
  case (1 TA)
  note trim_ta.simps[simp del]
  note trim_ta.simps[of TA, simp]
  show ?case
  proof (cases "finite (ta_states TA)")
    case False
    thus ?thesis by auto
  next
    case True
    note IH = 1[OF this refl]
    let ?R = "ta_only_reach TA"
    let ?PR = "ta_only_prod ?R"
    show ?thesis
    proof (cases "ta_states TA = ta_states ?PR")
      case True
      thus ?thesis by auto
    next
      case False
      note IH = IH[OF this]
      from ta_subset_trans[OF this ta_prod_reach_subset]
        False True show ?thesis by simp
    qed
  qed
qed

theorem trim_ta: "finite (ta_states TA) \<Longrightarrow> ta_trim (trim_ta TA)"
proof (induct TA rule: trim_ta.induct)
  case (1 TA)
  let ?R = "ta_only_reach TA"
  let ?PR = "ta_only_prod ?R"
  let ?P = "ta_only_prod TA"
  note trim_ta.simps[simp del]
  note trim_ta.simps[of TA, simp]
  from 1(2) have fin: "finite (ta_states TA)" by auto
  from ta_prod_reach_states[of TA] fin have "finite (ta_states ?PR)" by (rule finite_subset)
  note IH = 1(1)[OF fin refl _ this]
  show ?case
  proof (cases "ta_states TA = ta_states ?PR")
    case False
    note IH = IH[OF this]
    with fin False show ?thesis by simp
  next
    case True
    from ta_restrict_states[of ?R "{q. ta_productive ?R q}", folded True] 
    have "ta_states ?R = ta_states TA" using ta_restrict_states[of TA] by blast
    from ta_restrict_states_eq_imp_eq[OF this] have R: "?R = TA" .
    from ta_restrict_states_eq_imp_eq[OF True[unfolded R, symmetric]] have P: "?P = TA" .    
    from True have trim: "trim_ta TA = TA" by auto
    show ?thesis unfolding trim ta_trim_def
    proof (intro ballI)
      fix q
      assume q: "q \<in> ta_states TA"
      from ta_restrict_states_Q[of TA "{q. ta_productive TA q}", unfolded P] q
      have p: "ta_productive TA q" by auto
      from ta_restrict_states_Q[of TA "{q. ta_reachable TA q}", unfolded R] q
      have r: "ta_reachable TA q" by auto
      from p r show "ta_reachable TA q \<and> ta_productive TA q"
        by auto
    qed
  qed
qed

(* Proposition 1: every finite automaton can be turned into an 
  equivalent trim one *)
lemmas obtain_trimmed_ta = trim_ta trim_ta_lang ta_subset_det[OF trim_ta_subset]
thm obtain_trimmed_ta

declare trim_ta.simps[simp del]

subsection {* Theorem 3: Completeness of state-compatibility and state-coherence *}

lemma ta_trimE: assumes trim: "ta_trim TA"
  and rule: "(f qs \<rightarrow> q) \<in> ta_rules TA"
  and q': "q' \<in> set (q # qs)"
  shows "ta_reachable TA q' \<and> ta_productive TA q'"
proof -
  from q' rule have "q' \<in> ta_states TA" unfolding ta_states_def by (force simp: r_states_def)
  with trim show ?thesis unfolding ta_trim_def by auto
qed

(* we assume that there is some det. trim automaton TA and arbitrary TRS R such that 
   TA is closed under R *)         
locale ta_trim_det_closed = 
  fixes TA :: "('q,'f)ta"
  and R :: "('f,'q)trs"
  assumes det: "ta_det TA"
  and trim: "ta_trim TA"  
  and closed: "s \<in> ta_lang TA \<Longrightarrow> (s,t) \<in> rstep R \<Longrightarrow> t \<in> ta_lang TA"
begin

lemma [simp]: "ta_eps TA = {}" using det[unfolded ta_det_def] by auto

lemma ta_productiveE: assumes "ta_productive TA q"
  obtains q' C where "q' \<in> ta_res TA (C\<langle>Var q\<rangle>)" "q' \<in> ta_final TA" "ground_ctxt C"
  using assms[unfolded ta_productive_def] by auto

definition rel where "rel \<equiv> { (q,q'). \<exists> t t'. ground t \<and> ground t' \<and> q \<in> ta_res TA t \<and> (t,t') \<in> rstep R \<and> q' \<in> ta_res TA t'}"

lemma relE: assumes "(q,q') \<in> rel"
  obtains t t' where "ground t" "ground t'" "q \<in> ta_res TA t" "q' \<in> ta_res TA t'" "(t,t') \<in> rstep R"
  using assms[unfolded rel_def] by blast

lemma relI: "ground t \<Longrightarrow> ground t' \<Longrightarrow> q \<in> ta_res TA t \<Longrightarrow> q' \<in> ta_res TA t' \<Longrightarrow> (t,t') \<in> rstep R \<Longrightarrow> (q,q') \<in> rel"
  unfolding rel_def by blast

lemma compatible: "state_compatible TA rel R"
  unfolding state_compatible_def
proof (clarify)
  fix l r
  assume lr: "(l,r) \<in> R"
  show "rule_state_compatible TA rel (l,r)"
    unfolding rule_state_compatible_def
  proof (clarify)
    fix \<sigma> q 
    assume sigma: "\<sigma> ` vars_term l \<subseteq> ta_rhs_states TA"
      and ql: "q \<in> ta_res TA (map_vars \<sigma> l)"
    def \<tau> \<equiv> "\<lambda> x. (if x \<in> vars_term l then (SOME t. ground t \<and> \<sigma> x \<in> ta_res TA t) else Var q)"
    {
      fix x
      assume x: "x \<in> vars_term l"
      with sigma have "\<sigma> x \<in> ta_rhs_states TA" by auto
      with ta_rhs_states_subset_states[of TA] have mem: "\<sigma> x \<in> ta_states TA" by auto
      from trim[unfolded ta_trim_def, rule_format, OF mem]
      have "ta_reachable TA (\<sigma> x)" ..
      from ta_reachableE[OF this] have "\<exists> t. ground t \<and> \<sigma> x \<in> ta_res TA t" by blast
      from someI_ex[OF this] have "ground (\<tau> x) \<and> \<sigma> x \<in> ta_res TA (\<tau> x)" unfolding \<tau>_def using x by auto
    } note tau = this
    have "q \<in> ta_rhs_states TA"
    proof (cases l)
      case (Fun f ls)
      hence "is_Fun (map_vars \<sigma> l)" by simp
      from ta_rhs_states_res[OF this, of TA] ql show ?thesis by auto
    qed (insert ql sigma, auto)
    with ta_rhs_states_subset_states[of TA] have q: "q \<in> ta_states TA" by blast
    from trim[unfolded ta_trim_def, rule_format, OF this] have "ta_productive TA q" ..
    from ta_productiveE[OF this] obtain C p where 
      fin: "p \<in> ta_final TA" and pq: "p \<in> ta_res TA (C \<langle>Var q\<rangle>)" and C: "ground_ctxt C" .
    let ?t = "C \<langle> l \<cdot> \<tau> \<rangle>"
    have gt: "ground ?t" using C tau by auto
    have gl: "ground (l \<cdot> \<tau>)" using tau by auto
    have qlt: "q \<in> ta_res TA (l \<cdot> \<tau>)"
      by (rule ta_res_subst[OF _ ql], insert tau, auto)
    have "p \<in> ta_res TA ?t"
      by (rule ta_res_ctxt[OF qlt pq])
    hence tA: "?t \<in> ta_lang TA" unfolding ta_lang_def using fin gt by auto
    let ?t' = "C \<langle> r \<cdot> \<tau> \<rangle>"
    have "(?t, ?t') \<in> rstep R" using lr by auto
    from closed[OF tA this] have t'A: "?t' \<in> ta_lang TA" .
    from this[unfolded ta_lang_def] obtain p' where gr: "ground (r \<cdot> \<tau>)" and p': "p' \<in> ta_res TA ?t'" by auto
    from ta_res_ctxt_decompose[OF p'] obtain q' where q': "q' \<in> ta_res TA (r \<cdot> \<tau>)" by auto
    have lr: "(l \<cdot> \<tau>, r \<cdot> \<tau>) \<in> rstep R" using lr by auto
    have qq': "(q,q') \<in> rel" 
      by (rule relI[OF gl gr qlt q' lr])
    from ta_res_subst_det[OF det q']
      obtain \<tau>' where q': "q' \<in> ta_res TA (map_vars \<tau>' r)" and 
      tau': "\<And> x. x\<in>vars_term r \<Longrightarrow> \<tau>' x \<in> ta_res TA (\<tau> x)" by blast
    have "map_vars \<tau>' r = map_vars \<sigma> r" 
    proof (rule map_vars_vars_term)
      fix x
      assume xr: "x \<in> vars_term r"
      have xl: "x \<in> vars_term l"
      proof (rule ccontr)
        assume "x \<notin> vars_term l"
        hence "\<tau> x = Var q" unfolding \<tau>_def by auto
        with xr gr show False by auto
      qed      
      from tau'[OF xr] have 1: "\<tau>' x \<in> ta_res TA (\<tau> x)" by auto
      from tau[OF xl] have 2: "\<sigma> x \<in> ta_res TA (\<tau> x)" by auto
      from ta_detE[OF det 1 2] show "\<tau>' x = \<sigma> x" .
    qed
    hence q': "q' \<in> ta_res TA (map_vars \<sigma> r)" using q' by simp    
    thus "q \<in> rel\<inverse> `` ta_res TA (map_vars \<sigma> r)" using qq' by auto
  qed
qed

lemma coherent: "state_coherent TA rel"
  unfolding state_coherent_def
proof (intro conjI allI impI, rule)
  fix q'
  assume "q' \<in> rel `` ta_final TA"
  then obtain q where qq': "(q,q') \<in> rel" and qf: "q \<in> ta_final TA" by auto
  from relE[OF qq'] obtain t t' where 
    gt: "ground t" and gt': "ground t'" 
    and q: "q \<in> ta_res TA t" and q': "q' \<in> ta_res TA t'" 
    and step: "(t,t') \<in> rstep R" by auto
  from q gt qf have "t \<in> ta_lang TA" unfolding ta_lang_def by auto
  from closed[OF this step] have "t' \<in> ta_lang TA" .
  from this[unfolded ta_lang_def] obtain q'' where q'': "q'' \<in> ta_res TA t'" and qf'': "q'' \<in> ta_final TA" by auto
  from ta_detE[OF det q'' q'] qf''
  show "q' \<in> ta_final TA" by simp
next
  fix f qs q i qi'
  assume rule: "(f qs \<rightarrow> q) \<in> ta_rules TA" and
    i: "i < length qs" and
    rel: "(qs ! i, qi') \<in> rel"
  let ?n = "length qs"
  let ?qi = "qs ! i"
  note trim = ta_trimE[OF trim rule]
  from relE[OF rel] obtain ti ti' where
    gti: "ground ti" and gti': "ground ti'" 
    and qi: "?qi \<in> ta_res TA ti" 
    and qi': "qi' \<in> ta_res TA ti'" 
    and step: "(ti,ti') \<in> rstep R" by auto
  {
    fix j
    assume "j < ?n"
    hence "qs ! j \<in> set (q # qs)" by auto
    from trim[OF this] have "ta_reachable TA (qs ! j)" ..
    from ta_reachableE[OF this] have "\<exists> t. ground t \<and> qs ! j \<in> ta_res TA t" by blast
  }
  hence "\<forall> j. \<exists> t. j < ?n \<longrightarrow> ground t \<and> qs ! j \<in> ta_res TA t" by blast
  from choice[OF this] obtain tjs where tjs: "\<And> j. j < ?n \<Longrightarrow> ground (tjs j) \<and> qs ! j \<in> ta_res TA (tjs j)" by blast
  let ?t_gen = "\<lambda> t. Fun f (map (\<lambda> j. if i = j then t else tjs j) [0 ..< ?n])"
  let ?t = "?t_gen ti"
  let ?t' = "?t_gen ti'"
  from trim[of q] have "ta_productive TA q" by auto
  from ta_productiveE[OF this]
    obtain p C where qp: "p \<in> ta_res TA (C\<langle>Var q\<rangle>)" 
    and p: "p \<in> ta_final TA" and gC: "ground_ctxt C" .
  {
    fix t :: "('f,'q)term"
    assume "ground t"
    hence "ground (?t_gen t)" 
      using tjs by auto
  } note g = this
  from g[OF gti] have gCt: "ground (C\<langle>?t\<rangle>)" using gC by auto
  from g[OF gti'] have gCt': "ground (C\<langle>?t'\<rangle>)" using gC by auto
  have qt: "q \<in> ta_res TA ?t"
    by (simp, intro exI conjI, rule rule, insert qi tjs, auto)
  have pCt: "p \<in> ta_res TA (C \<langle> ?t \<rangle>)"
    by (rule ta_res_ctxt[OF qt qp])
  from p pCt gCt have tA: "C \<langle> ?t \<rangle> \<in> ta_lang TA" by (auto simp: ta_lang_def)
  let ?C = "ctxt_of_pos_term (<i>) ?t"
  from rstep.ctxt[OF step, of ?C] have "(?C \<langle> ti \<rangle>, ?C \<langle> ti' \<rangle>) \<in> rstep R" .
  also have "?C \<langle> ti \<rangle> = ?t" unfolding map_upt_split[OF i] by auto
  also have "?C \<langle> ti' \<rangle> = ?t'" unfolding map_upt_split[OF i] by auto
  finally have step: "(?t,?t') \<in> rstep R" .
  from closed[OF tA rstep.ctxt[OF step, of C]] have "C\<langle>?t'\<rangle> \<in> ta_lang TA" .
  then obtain p' where pt': "p' \<in> ta_res TA (C\<langle>?t'\<rangle>)" and p': "p' \<in> ta_final TA" by (auto simp: ta_lang_def)
  from ta_res_ctxt_decompose[OF pt'] obtain q' where qt': "q' \<in> ta_res TA ?t'" and pq': "p' \<in> ta_res TA (C \<langle>Var q'\<rangle>)" by auto
  let ?tj_gen = "\<lambda> t j. if i = j then t else tjs j"
  let ?ti = "?tj_gen ti"
  let ?ti' = "?tj_gen ti'"
  from qt'
  obtain qs' where rule: "(f qs' \<rightarrow> q') \<in> ta_rules TA"
    and len: "length qs' = length qs"
    and qj: "\<And> j. j<length qs \<Longrightarrow> qs' ! j \<in> ta_res TA (?ti' j)" by auto
  have "qs' = qs[i := qi']"
  proof (rule nth_equalityI, simp add: len, intro allI impI)
    fix j
    assume j': "j < length qs'"
    hence j: "j < length qs" by (simp add: len)
    from qj[OF this] have qj: "qs' ! j \<in> ta_res TA (?ti' j)" .
    show "qs' ! j = qs[i := qi'] ! j"
    proof (cases "j = i")
      case True
      with qj have "qs' ! j \<in> ta_res TA ti'" by simp
      from ta_detE[OF det this qi'] True show ?thesis using j by auto
    next
      case False
      with qj have "qs' ! j \<in> ta_res TA (tjs j)" by auto
      from ta_detE[OF det this] tjs[OF j] False show ?thesis using j by auto
    qed
  qed
  with rule have "(f qs[i := qi'] \<rightarrow> q') \<in> ta_rules TA" by auto
  moreover
  from relI[OF g[OF gti] g[OF gti'] qt qt' step]
  have "(q,q') \<in> rel" .
  ultimately show "\<exists>q'. (f qs[i := qi'] \<rightarrow> q') \<in> ta_rules TA \<and> (q, q') \<in> rel" by blast
qed
end

(* Theorem 3 in paper *)
theorem ta_trim_det_closed:
  fixes TA :: "('q,'f)ta"
  and R :: "('f,'q)trs"
  assumes det: "ta_det TA"
  and trim: "ta_trim TA"  
  and closed: "\<And> s t. s \<in> ta_lang TA \<Longrightarrow> (s,t) \<in> rstep R \<Longrightarrow> t \<in> ta_lang TA"
  shows "\<exists> rel. state_compatible TA rel R \<and> state_coherent TA rel"
proof -
  interpret ta_trim_det_closed TA R
    by (unfold_locales, rule det, rule trim, rule closed)
  show ?thesis
    by (intro exI conjI, rule compatible, rule coherent)
qed

(* First part of Corollary 1 in paper *)
corollary closed_iff_compatible_and_coherent:
  fixes TA :: "('q,'f)ta" and R :: "('f,'q)trs"
  assumes determinizer: "\<And> TA :: ('q,'f)ta. finite (ta_states TA) \<Longrightarrow>
    finite (ta_states (determ TA)) \<and> ta_lang (determ TA) = ta_lang TA \<and> ta_det (determ TA)"
  and wf: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and fin: "finite (ta_states TA)"
  shows "(\<forall> s t. s \<in> ta_lang TA \<and> (s,t) \<in> rstep R \<longrightarrow> t \<in> ta_lang TA) \<longleftrightarrow>
    (\<exists> rel. state_compatible (trim_ta (determ TA)) rel R \<and> state_coherent (trim_ta (determ TA)) rel)" 
    (is "?closed = ?compat_coherent")
proof -
  def TA' \<equiv> "trim_ta (determ TA)"
  from determinizer[OF fin] have d: "ta_lang (determ TA) = ta_lang TA" and det: "ta_det (determ TA)" by auto
  have lang: "ta_lang TA' = ta_lang TA" unfolding TA'_def by (simp add: d trim_ta_lang)
  from ta_subset_det[OF trim_ta_subset det] have det: "ta_det TA'" unfolding TA'_def .
  show ?thesis
  proof
    assume ?compat_coherent
    then obtain rel where compat: "state_compatible TA' rel R" "state_coherent TA' rel" 
      unfolding TA'_def by blast
    from det have eps: "ta_eps TA' = {}" unfolding ta_det_def by auto
    from state_compatible[OF compat disjI2[OF det] wf subset_refl]
    show ?closed unfolding lang by blast
  next
    assume ?closed
    have "finite (ta_states (determ TA))" using determinizer[OF fin] ..
    from trim_ta[OF this] have trim: "ta_trim TA'" unfolding TA'_def .
    from ta_trim_det_closed[OF det trim, of R] `?closed`
    show ?compat_coherent unfolding lang[symmetric] TA'_def[symmetric] by blast
  qed
qed

subsection {* Computing smallest relations such that
  automaton is state-coherent and state-compatible (algorithm of Section 3.3) *}
locale det_automaton = fixes 
  TA :: "('q,'f)ta" and
  R  :: "('f,'q)trs" 
assumes
  det: "ta_det TA"  
  and fin: "finite (ta_states TA)"
  and wf: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
begin

inductive require_relation :: "('q \<times> 'q)set \<Rightarrow> 'q \<Rightarrow> 'q option \<Rightarrow> bool" for rel :: "('q \<times> 'q)set" where
  compat: "(l,r) \<in> R \<Longrightarrow> q \<in> ta_res TA (map_vars \<sigma> l) \<Longrightarrow> q' \<in> ta_res TA (map_vars \<sigma> r) \<Longrightarrow> 
    \<sigma> ` vars_term l \<subseteq> ta_rhs_states TA \<Longrightarrow> 
    require_relation rel q (Some q')" 
| no_compat: "(l,r) \<in> R \<Longrightarrow> q \<in> ta_res TA (map_vars \<sigma> l) \<Longrightarrow> ta_res TA (map_vars \<sigma> r) = {} \<Longrightarrow> 
    \<sigma> ` vars_term l \<subseteq> ta_rhs_states TA \<Longrightarrow>
    require_relation rel q None" 
| coherence: "(f qs \<rightarrow> q) \<in> ta_rules TA \<Longrightarrow> i < length qs \<Longrightarrow> (qs ! i, qi') \<in> rel 
    \<Longrightarrow> (f (qs[i := qi']) \<rightarrow> q') \<in> ta_rules TA \<Longrightarrow> 
    require_relation rel q (Some q')" 
| no_coherence: "(f qs \<rightarrow> q) \<in> ta_rules TA \<Longrightarrow> i < length qs \<Longrightarrow> (qs ! i, qi') \<in> rel 
    \<Longrightarrow> \<not> (\<exists> q'. (f (qs[i := qi']) \<rightarrow> q') \<in> ta_rules TA) \<Longrightarrow> 
    require_relation rel q None" 

lemma require_relation: assumes sub: "rel \<subseteq> rel'" 
  shows "require_relation rel q sq' \<Longrightarrow> state_coherent TA rel' \<Longrightarrow> state_compatible TA rel' R \<Longrightarrow> 
   \<exists> q'. sq' = Some q' \<and> (q,q') \<in> rel'"
proof (induct rule: require_relation.induct)
  case (no_compat l r q \<sigma>)
  from no_compat(2) have "funas_term l \<subseteq> ta_syms TA"
    by (metis funas_term_map_vars ta_syms_res)
  with no_compat(1,6)[unfolded state_compatible_def]
  have "rule_state_compatible TA rel' (l, r)" by auto
  from this[unfolded rule_state_compatible_def split] no_compat(4) no_compat(2,3)
  show ?case by blast
next
  case (compat l r q \<sigma> q')
  from compat(2) have "funas_term l \<subseteq> ta_syms TA"
    by (metis funas_term_map_vars ta_syms_res)
  with compat(1,6)[unfolded state_compatible_def]
  have "rule_state_compatible TA rel' (l, r)" by auto
  from this[unfolded rule_state_compatible_def split, rule_format, OF compat(4)] compat(2) 
  obtain q'' where "q'' \<in> ta_res TA (map_vars \<sigma> r)" "(q,q'') \<in> rel'" by blast
  from ta_detE[OF det compat(3) this(1)] this(2) have "(q,q') \<in> rel'" by auto
  thus ?case by auto
next
  case (no_coherence f qs q i qi')
  with sub have mem: "(qs ! i, qi') \<in> rel'" by auto
  from no_coherence(5) have "state_coherent TA rel'" by auto
  from this[unfolded state_coherent_def, THEN conjunct2, rule_format, OF no_coherence(1-2) mem]
    no_coherence(4) show ?case by blast
next
  case (coherence f qs q i qi' q')
  with sub have mem: "(qs ! i, qi') \<in> rel'" by auto
  from coherence(5) have "state_coherent TA rel'" by auto
  from this[unfolded state_coherent_def, THEN conjunct2, rule_format, OF coherence(1-2) mem]
  obtain q'' where "(f qs[i := qi'] \<rightarrow> q'') \<in> ta_rules TA" "(q,q'') \<in> rel'" by auto
  with det[unfolded ta_det_def] coherence(4) show ?case by blast
qed

inductive_set generate_rel :: "(('q \<times> 'q)set option \<times> ('q \<times> 'q)set option)set" where
  fail: "require_relation rel q None \<Longrightarrow> (Some rel, None) \<in> generate_rel"
| fail2: "require_relation rel q (Some q') \<Longrightarrow> q \<in> ta_final TA \<Longrightarrow> q' \<notin> ta_final TA \<Longrightarrow> (Some rel, None) \<in> generate_rel"
| iterate: "require_relation rel q (Some q') \<Longrightarrow> (q,q') \<notin> rel \<Longrightarrow> (q \<in> ta_final TA \<longrightarrow> q' \<in> ta_final TA) \<Longrightarrow>
  (Some rel, Some (insert (q,q') rel)) \<in> generate_rel"

lemma generate_rel_subset: "(rel,srel') \<in> generate_rel \<Longrightarrow> the rel \<subseteq> rel'' 
  \<Longrightarrow> state_coherent TA rel'' \<Longrightarrow> state_compatible TA rel'' R 
  \<Longrightarrow> \<exists> rel'. srel' = Some rel' \<and> rel' \<subseteq> rel''"
proof (induct rule: generate_rel.induct)
  case (fail rel q)
  note fail = fail[unfolded the.simps]
  from require_relation[OF fail(2) fail(1,3,4)]
  show ?case by blast
next
  case (fail2 rel q q')
  note fail2 = fail2[unfolded the.simps]
  from require_relation[OF fail2(4) fail2(1,5,6)] fail2(2,3)
  fail2(5)[unfolded state_coherent_def, THEN conjunct1] show ?case by blast
next
  case (iterate rel q q')
  note iterate = iterate[unfolded the.simps]
  from require_relation[OF iterate(4,1,5,6)] iterate(4)
  show ?case by auto
qed

lemma generate_rel_complete: assumes steps: "(Some {},srel) \<in> generate_rel^*"
  and coh: "state_coherent TA rel'"
  and comp: "state_compatible TA rel' R"
  shows "\<exists> rel. srel = Some rel \<and> rel \<subseteq> rel'"
proof -
  {
    fix sr
    have "(sr, srel) \<in> generate_rel^* \<Longrightarrow> sr \<noteq> None \<Longrightarrow> the sr \<subseteq> rel' \<Longrightarrow> \<exists> rel. srel = Some rel \<and> rel \<subseteq> rel'"
    proof (induct rule: rtrancl_induct)
      case base
      thus ?case by (cases sr, auto)
    next
      case (step sr' sr'')
      from step(3)[OF step(4-5)] obtain r' where sr': "sr' = Some r'" and sub: "r' \<subseteq> rel'" by auto
      from generate_rel_subset[OF step(2), unfolded sr' the.simps, OF sub coh comp]
      show ?case .
    qed
  }
  from this[OF steps] show ?thesis by auto
qed
  
lemma generate_rel_sound: assumes nf: "(Some {}, Some rel) \<in> generate_rel^!"
  shows "state_coherent TA rel \<and> state_compatible TA rel R"
proof (rule ccontr)
  from nf have NF: "Some rel \<in> NF (generate_rel)" by auto
  {
    fix srel
    assume "(Some {}, srel) \<in> generate_rel^*"
    hence "srel \<noteq> None \<Longrightarrow> the (srel) `` ta_final TA \<subseteq> ta_final TA"
    proof (induct rule: rtrancl_induct)
      case (step srel srel')
      from step(2) obtain rel where srel: "srel = Some rel" by (induct, auto)
      from step(4) obtain rel' where srel': "srel' = Some rel'" by (cases, auto)
      from step(2) srel srel' have "\<exists> q q'. the (srel') = insert (q,q') (the srel) \<and> (q \<in> ta_final TA \<longrightarrow> q' \<in> ta_final TA)"
        by (induct rule: generate_rel.induct, auto)
      with srel srel' step(3) show ?case by auto
    qed auto
  }
  from this[of "Some rel"] nf
  have rel_fin: "rel `` ta_final TA \<subseteq> ta_final TA" by auto
  hence fin_imp: "\<And> q q'. (q,q') \<in> rel \<Longrightarrow> q \<in> ta_final TA \<longrightarrow> q' \<in> ta_final TA" by blast
  assume "\<not> ?thesis"
  hence "\<not> state_coherent TA rel \<or> \<not> state_compatible TA rel R" by blast
  thus False
  proof
    assume "\<not> state_compatible TA rel R"
    from this[unfolded state_compatible_def]
    obtain l r where 
      lr: "(l, r) \<in> R" and fl: "funas_term l \<subseteq> ta_syms TA" 
      and not: "\<not> rule_state_compatible TA rel (l, r)" by blast
    from not[unfolded rule_state_compatible_def split] obtain \<sigma> where
      vars: "\<sigma> ` vars_term l \<subseteq> ta_rhs_states TA" 
      and not: "\<not> ta_res TA (map_vars \<sigma> l) \<subseteq> rel\<inverse> `` ta_res TA (map_vars \<sigma> r)" by blast
    from not obtain q where q: "q \<in> ta_res TA (map_vars \<sigma> l)" 
      and not: "\<not> (\<exists> q'. q' \<in> ta_res TA (map_vars \<sigma> r) \<and> (q,q') \<in> rel)" by blast
    show False
    proof (cases "\<exists> q'. q' \<in> ta_res TA (map_vars \<sigma> r)")
      case True
      then obtain q' where q': "q' \<in> ta_res TA (map_vars \<sigma> r)" by auto
      with not have qq': "(q,q') \<notin> rel" by blast
      from require_relation.compat[OF lr q q' vars] have "require_relation rel q (Some q')" .
      from generate_rel.iterate[OF this qq'] generate_rel.fail2[OF this] NF show False by blast
    next
      case False
      hence "ta_res TA (map_vars \<sigma> r) = {}" by blast
      from require_relation.no_compat[OF lr q this vars] have "require_relation rel q None" .
      from generate_rel.fail[OF this] NF show False by blast
    qed
  next
    assume "\<not> state_coherent TA rel"
    from this[unfolded state_coherent_def] rel_fin
    obtain f qs q i qi
    where rule: "(f qs \<rightarrow> q) \<in> ta_rules TA"
          and i: "i < length qs" and rel: "(qs ! i, qi) \<in> rel" 
          and not: "\<not> (\<exists> q'. (f qs[i := qi] \<rightarrow> q') \<in> ta_rules TA \<and> (q, q') \<in> rel)"
      by blast
    show False
    proof (cases "\<exists> q'. (f qs[i := qi] \<rightarrow> q') \<in> ta_rules TA")
      case True
      then obtain q' where rule': "(f qs[i := qi] \<rightarrow> q') \<in> ta_rules TA" by blast
      with not have qq': "(q,q') \<notin> rel" by auto
      from require_relation.coherence[OF rule i rel rule'] have "require_relation rel q (Some q')" .
      from generate_rel.iterate[OF this qq'] generate_rel.fail2[OF this]
        NF show False by blast
    next
      case False
      from generate_rel.fail[OF require_relation.no_coherence[OF rule i rel False]] NF
      show False by blast
    qed
  qed
qed

(* termination of generation, obvious, but tedious *)
lemma SN_generate_rel: "SN generate_rel" 
proof 
  fix f
  assume "\<forall> i. (f i, f (Suc i)) \<in> generate_rel"
  hence rel: "\<And> i. (f i, f (Suc i)) \<in> generate_rel" ..
  {
    fix i
    from rel[of i] have "\<exists> r. f i = Some r" by (induct, auto)
  }
  hence "\<forall> i. \<exists> r. f i = Some r" ..
  from choice[OF this] obtain r where fr: "\<And> i. f i = Some (r i)" by blast
  def S \<equiv> "ta_states TA \<times> ta_states TA"
  def T \<equiv> "\<lambda> i. S - r i"
  {
    fix i
    from rel[of i] fr[of i] fr[of "Suc i"] have "\<exists> q q'. require_relation (r i) q (Some q') \<and>
      (q,q') \<notin> r i \<and> r (Suc i) = insert (q,q') (r i)"
      by (induct, auto)
    then obtain q q' where req: "require_relation (r i) q (Some q')"
      and nmem: "(q,q') \<notin> r i" and rsi: "r (Suc i) = insert (q,q') (r i)" by auto
    {
      fix q :: 'q and l :: "('f,'q)term" and t \<sigma>
      assume *: "q \<in> ta_res TA (map_vars \<sigma> t)" "\<sigma> ` vars_term l \<subseteq> ta_rhs_states TA"
         "vars_term t \<subseteq> vars_term l"
      have "q \<in> ta_rhs_states TA"
      proof (cases "is_Fun t")
        case True
        hence "is_Fun (map_vars \<sigma> t)" by auto
        from ta_rhs_states_res[OF this, of TA] *(1) show ?thesis by blast
      next
        case False
        then obtain q' where t: "t = Var q'" by auto
        with * have "\<sigma> q' \<in> ta_rhs_states TA" by auto
        with *(1) det[unfolded ta_det_def] show ?thesis unfolding t by auto
      qed
      hence "q \<in> ta_states TA" 
        by (metis (full_types) contra_subsetD ta_rhs_states_subset_states)
    }
    note in_q = this
    def sq' \<equiv> "Some q'" 
    from req have "require_relation (r i) q sq'" and "sq' = Some q'" unfolding sq'_def by auto    
    hence "{q,the sq'} \<subseteq> ta_states TA"
    proof (induct)
      case compat
      from in_q[OF compat(2,4)] in_q[OF compat(3,4) wf[OF compat(1)]] show ?case by auto
    next
      case (coherence f qs q i qi' q')
      have "q \<in> r_states (f qs \<rightarrow> q)" unfolding r_states_def by auto
      moreover have "q' \<in> r_states (f qs[i := qi'] \<rightarrow> q')" unfolding r_states_def by auto
      ultimately show ?case using coherence(1,4)
        by (auto simp: ta_states_def)
    qed auto
    with sq'_def have ta: "{q,q'} \<subseteq> ta_states TA" by auto
    hence "(q,q') \<in> S" unfolding S_def by blast
    with nmem have "S - r i \<supset> S - r (Suc i)" unfolding rsi by blast
    hence "(S - r (Suc i), S - r i) \<in> finite_psubset" unfolding in_finite_psubset
    proof (rule conjI)
      show "finite (S - r i)" unfolding S_def using fin by auto
    qed
    hence "(T (Suc i), T i) \<in> finite_psubset" unfolding T_def .
  } note T = this
  hence "\<exists>T. \<forall>i. (T (Suc i), T i) \<in> (finite_psubset :: (('q \<times> 'q) set \<times> ('q \<times> 'q) set) set)" by blast
  with wf_finite_psubset[unfolded wf_iff_no_infinite_down_chain] show False by blast
qed

(* decision procedure for coherence and compatibility: just compute arbitrary normal form
  w.r.t. generate_rel, and then check whether relation was produced *)
definition "decide_coherent_compatible \<equiv> case (SOME srel. (Some {}, srel) \<in> generate_rel^!) of
  None \<Rightarrow> False
| Some rel \<Rightarrow> True"

lemma decide_coherent_compatible: "decide_coherent_compatible = 
  (\<exists> rel. state_compatible TA rel R \<and> state_coherent TA rel)"
proof -
  let ?P = "\<lambda> srel. (Some {}, srel) \<in> generate_rel^!"
  def srel \<equiv> "SOME srel. ?P srel"
  from SN_generate_rel have "SN_on generate_rel {Some {}}" by fast
  from SN_on_imp_normalizability[OF this] obtain b where P: "?P b" by blast
  have P: "?P srel" 
    unfolding srel_def
    by (rule someI, rule P)
  hence steps: "(Some {}, srel) \<in> generate_rel^*" by auto
  note d = decide_coherent_compatible_def srel_def[symmetric]
  show ?thesis
  proof
    assume "decide_coherent_compatible"
    from this[unfolded d] obtain rel where srel: "srel = Some rel" by (cases srel, auto)
    from generate_rel_sound[OF P[unfolded srel]] show "(\<exists> rel. state_compatible TA rel R \<and> state_coherent TA rel)" by blast
  next
    assume "(\<exists> rel. state_compatible TA rel R \<and> state_coherent TA rel)"
    then obtain rel where compat: "state_compatible TA rel R" and coh: "state_coherent TA rel" by blast
    from generate_rel_complete[OF steps coh compat] show "decide_coherent_compatible"
      unfolding d by auto
  qed
qed
end
  
subsection {* Abstract algorithms to check compatibility for a single rule and containment *}

(* ta_match computes (a usually small) subset of those state-substitutions
  that have to be considered for some left-hand side *)
fun ta_match :: "('q,'f)ta \<Rightarrow> ('f,'v)term \<Rightarrow> 'q \<Rightarrow> ('v \<Rightarrow> 'q) set"
  where "ta_match TA (Var x) q = { (\<lambda> _. q') | q'. (q',q) \<in> (ta_eps TA)^*}"
   | "ta_match TA (Fun f ts) q = { fun_merge \<sigma>s (map vars_term ts) | \<sigma>s qs q'. 
          (f qs \<rightarrow> q') \<in> ta_rules TA \<and> 
          (q',q) \<in> (ta_eps TA)^* \<and> 
          length qs = length ts \<and>
          length \<sigma>s = length ts \<and>
          (\<forall> i < length ts. \<sigma>s ! i \<in> ta_match TA (ts ! i) (qs ! i)) }"

(* whenever a state is reachable via some substitution \<tau>, then
   there is an equivalent one (\<sigma>) that is computed by ta_match *)
lemma ta_match: 
  assumes "q \<in> ta_res TA (map_vars \<tau> t)"
                shows "\<exists> \<sigma> \<in> ta_match TA t q. (\<forall> x \<in> vars_term t. \<tau> x = \<sigma> x)"
  using assms
proof (induct t arbitrary: q)
  case (Var x)
  thus ?case by auto
next
  case (Fun f ts)
  from Fun(2)
  obtain q' qs where rule: "(f qs \<rightarrow> q') \<in> ta_rules TA"
    and eps: "(q',q) \<in> (ta_eps TA)^*"
    and len: "length qs = length ts"
    and rec: "\<And> i. i < length ts \<Longrightarrow> qs ! i \<in> ta_res TA (map_vars \<tau> (ts ! i))"
    by auto
  let ?p = "\<lambda> i \<sigma>. \<sigma> \<in> ta_match TA (ts ! i) (qs ! i) \<and> (\<forall> x \<in> vars_term (ts ! i). \<tau> x = \<sigma> x)" 
  {
    fix i
    assume i: "i < length ts"
    hence ti: "ts ! i \<in> set ts" by auto
    from Fun(1)[OF ti rec[OF i]] 
    have "\<exists> \<sigma>. ?p i \<sigma>" by blast
  }
  hence "\<forall> i. \<exists> \<sigma>. i < length ts \<longrightarrow> ?p i \<sigma>" by auto
  from choice[OF this] obtain \<sigma>s where \<sigma>s: "\<And> i. i < length ts \<Longrightarrow> ?p i (\<sigma>s i)" by auto
  let ?\<sigma>s = "map \<sigma>s [0 ..< length ts]"
  let ?xs = "map vars_term ts"
  obtain \<sigma> where \<sigma>: "\<sigma> \<equiv> fun_merge ?\<sigma>s ?xs" by auto
  have mem: "\<sigma> \<in> ta_match TA (Fun f ts) q"
    unfolding ta_match.simps
    by (rule, intro exI conjI, unfold \<sigma>, 
    rule HOL.refl,
    rule rule,
    insert len eps \<sigma>s, auto)
  show ?case
  proof (rule bexI[OF _ mem], intro ballI)
    fix x 
    assume "x \<in> vars_term (Fun f ts)"
    then obtain t where "t \<in> set ts" and x: "x \<in> vars_term t" by auto
    then obtain i where i: "i < length ts" and t: "t = ts ! i" 
      unfolding set_conv_nth by auto
    from i x[unfolded t] have x: "x \<in> (map vars_term ts) ! i" by auto
    have "\<sigma> x = (?\<sigma>s ! i) x" unfolding \<sigma> 
    proof (rule fun_merge[OF _ x, of ?\<sigma>s, unfolded length_map, OF i])
      fix i j x
      assume i: "i < length ts" and j: "j < length ts" and xi: "x \<in> ?xs ! i" and xj: "x \<in> ?xs ! j"
      have "(?\<sigma>s ! i) x = \<tau> x" using \<sigma>s[OF i] xi i by auto 
      also have "... = (?\<sigma>s ! j) x" using \<sigma>s[OF j] xj j by auto 
      finally show "(?\<sigma>s ! i) x = (?\<sigma>s ! j) x" .
    qed
    also have "... = \<tau> x"
      using \<sigma>s[OF i] x i by auto
    finally show "\<tau> x = \<sigma> x" by simp
  qed
qed

(* efficient algorithm to check compatibility *)
(* one might examine whether \<sigma> in ta_match is always able to produce q,
   then one can ignore the precondition "q \<in> ta_res TA (Fun f ls \<cdot> \<sigma>)" *)
fun rule_state_compatible_eff :: "('q,'f)ta \<Rightarrow> 'q rel \<Rightarrow> ('f,'v)rule \<Rightarrow> bool"
where "rule_state_compatible_eff TA rel (Fun f ls,r) = (\<forall> \<sigma> qs q.
  (f qs \<rightarrow> q) \<in> ta_rules TA \<and>
  length qs = length ls \<and>
  \<sigma> \<in> ta_match TA (Fun f ls) q \<and>
  q \<in> ta_res TA (map_vars \<sigma> (Fun f ls)) \<longrightarrow>
  (\<exists> q''. (q,q'') \<in> rel \<and> q'' \<in> ta_res TA (map_vars \<sigma> r)))"
 | "rule_state_compatible_eff TA rel (Var x,r) = 
   (\<forall> q \<in> ta_rhs_states TA. (\<exists> q'. (q,q') \<in> rel \<and> q' \<in> ta_res TA (map_vars (\<lambda> _.q) r)))"


lemma rule_state_compatible_eff: 
  assumes var: "vars_term l \<supseteq> vars_term r"
  and no_eps_or_rel_id: "ta_eps TA = {} \<or> rel = Id"
  and compat: "rule_state_compatible_eff TA rel (l,r)"
  shows "rule_state_compatible TA rel (l,r)"
  unfolding rule_state_compatible_def split
proof (intro allI impI)
  fix \<tau>
  assume \<tau>: "\<tau> ` vars_term l \<subseteq> ta_rhs_states TA"
  show "ta_res TA (map_vars \<tau> l) \<subseteq> rel^-1 `` ta_res TA (map_vars \<tau> r)" (is "?rl \<subseteq> ?rr")
  proof (cases l)
    case (Fun f ls) note l = this
    show ?thesis
    proof
      fix q'
      assume q': "q' \<in> ?rl"
      then obtain qs q where rule: "(f qs \<rightarrow> q) \<in> ta_rules TA" and len: "length qs = length ls" 
        and eps: "(q,q') \<in> (ta_eps TA)^*" and q: "q \<in> ?rl" unfolding l
        by auto
      from ta_match[OF q] obtain \<sigma> where \<sigma>: "\<sigma> \<in> ta_match TA l q" and \<tau>: "\<And> x. x \<in> vars_term l \<Longrightarrow> \<tau> x = \<sigma> x" by auto
      have id: "map_vars \<tau> l = map_vars \<sigma> l"
        by (rule map_vars_vars_term, insert \<tau>, auto)
      from compat[unfolded l rule_state_compatible_eff.simps, THEN spec[of _ \<sigma>], THEN spec[of _ qs], THEN spec[of _ q]] 
rule len \<sigma> l q[unfolded id] 
      obtain q'' where rel: "(q,q'') \<in> rel" and  q'': "q'' \<in> ta_res TA (map_vars \<sigma> r)" by auto
      have "map_vars \<sigma> r = map_vars \<tau> r"
      proof (rule map_vars_vars_term)
        fix x
        assume "x \<in> vars_term r"
        with var have "x \<in> vars_term l" by auto
        with \<tau> show "\<sigma> x = \<tau> x" by simp
      qed
      with q'' have q'': "q'' \<in> ta_res TA (map_vars \<tau> r)" by simp
      from no_eps_or_rel_id show "q' \<in> ?rr"
      proof
        assume "ta_eps TA = {}"
        with eps have qq': "q = q'" by auto
        with rel q'' show ?thesis  by auto
      next
        assume "rel = Id"
        with rel have qq'': "q = q''" by auto
        from ta_res_eps[OF q'' eps[unfolded qq'']] show "q' \<in> ?rr" using `rel = Id` by auto
      qed
    qed
  next
    case (Var x)
    with \<tau> have \<tau>: "\<tau> x \<in> ta_rhs_states TA" by auto
    let ?\<tau> = "\<lambda> _ . \<tau> x"
    from compat[unfolded Var rule_state_compatible_eff.simps, THEN bspec[OF _ \<tau>]] obtain q
    where \<tau>q: "(\<tau> x,q) \<in> rel" and q: "q \<in> ta_res TA (map_vars ?\<tau> r)" by auto
    have "map_vars ?\<tau> r = map_vars \<tau> r"
    proof (rule map_vars_vars_term)
      fix y
      assume "y \<in> vars_term r"
      with var have "y \<in> vars_term l" by auto
      with Var have "y = x" by simp
      thus "\<tau> x = \<tau> y" by simp
    qed
    with q have q: "q \<in> ta_res TA (map_vars \<tau> r)" by simp
    show ?thesis
    proof
      fix q'
      assume "q' \<in> ?rl" 
      hence eps: "(\<tau> x,q') \<in> (ta_eps TA)^*" unfolding Var by simp
      from no_eps_or_rel_id show "q' \<in> ?rr"
      proof
        assume "ta_eps TA = {}"
        with eps have "\<tau> x = q'" by auto
        with q \<tau>q show ?thesis by auto
      next
        assume id: "rel = Id"
        with \<tau>q have "\<tau> x = q" by auto 
        from ta_res_eps[OF q eps[unfolded this]] id
        show ?thesis by auto
      qed
    qed    
  qed
qed

(* sufficient criterion to check whether automaton accepts T(F) where F is a subset of
   the signature *) 
definition ta_contains :: "('f \<times> nat)set \<Rightarrow> ('q,'f)ta \<Rightarrow> 'q set \<Rightarrow> bool"
  where "ta_contains F TA Q \<equiv> Q \<subseteq> ta_final TA
  \<and> (\<forall> f qs. (f,length qs) \<in> F \<and> set qs \<subseteq> Q \<longrightarrow> (\<exists> q q'. (f qs \<rightarrow> q) \<in> ta_rules TA \<and> q' \<in> Q \<and> (q,q') \<in> (ta_eps TA)^*))"

lemma ta_contains: 
  assumes contain: "ta_contains F TA qs"
  shows "({t :: ('f,'q)term. funas_term t \<subseteq> F}) \<inter> {t. ground t} \<subseteq> ta_lang TA"
proof (clarify)
  fix t :: "('f,'q)term"
  note contain = contain[unfolded ta_contains_def, simplified]
  assume "funas_term t \<subseteq> F" and g: "ground t"
  hence "\<exists> q \<in> qs. q \<in> ta_res TA t"
  proof (induct t)
    case (Var x) thus ?case by simp
  next
    case (Fun f ts)
    {
      fix i
      assume "i < length ts"
      hence t: "ts ! i \<in> set ts" by auto
      from Fun(3) t have gt: "ground (ts ! i)" by auto
      from Fun(2) t have fs: "funas_term (ts ! i) \<subseteq> F" by auto
      note Fun(1)[OF t fs gt] 
    } hence "\<forall> i. \<exists> q. i < length ts \<longrightarrow> q \<in> qs \<and> q \<in> ta_res TA (ts ! i)" by auto
    from choice[OF this] obtain qs' where qs'res: "\<And>i. i < length ts \<Longrightarrow> qs' i \<in> qs \<and> qs' i \<in> ta_res TA (ts ! i)" by auto
    let ?qs = "map qs' [0 ..< length ts]"
    have qs': "\<And> i. i < length ts \<Longrightarrow>  ?qs ! i = qs' i" by auto
    with qs'res have qs'res: "\<forall> i < length ts. ?qs ! i \<in> qs \<and> ?qs ! i \<in> ta_res TA (ts ! i)" by auto
    from contain[THEN conjunct2, THEN spec[of _ f], THEN spec[of _ ?qs]] Fun(2) qs'res 
    obtain q' q where q: "q \<in> qs" and q': "(q',q) \<in> (ta_eps TA)^*" and rule: "(f ?qs \<rightarrow> q') \<in> ta_rules TA" by force
    show ?case unfolding ta_res.simps
      by (rule bexI[OF _ q], rule, rule exI[of _ q], rule exI[of _ q'], rule exI[of _ ?qs], insert rule qs'res, simp add: q')
  qed
  then obtain q where q1: "q \<in> qs" and q2: "q \<in> ta_res TA t" by auto
  from contain[THEN conjunct1] q1 have q1: "q \<in> ta_final TA"
    by auto
  show "t \<in> ta_lang TA" unfolding ta_lang_def using q1 q2 g by auto
qed  


end
